from __future__ import annotations

import http.server as server_base
import json
import logging
import multiprocessing
import sys
import time
import urllib.parse

import libtbx.phil
from cctbx import uctbx
from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx.introspection import number_of_processors

from dials.algorithms.indexing import indexer
from dials.algorithms.integration.integrator import create_integrator
from dials.algorithms.profile_model.factory import ProfileModelFactory
from dials.algorithms.spot_finding import per_image_analysis
from dials.array_family import flex
from dials.command_line.find_spots import phil_scope as find_spots_phil_scope
from dials.command_line.index import phil_scope as index_phil_scope
from dials.command_line.integrate import phil_scope as integrate_phil_scope
from dials.util import Sorry, show_mail_handle_errors
from dials.util.options import ArgumentParser

logger = logging.getLogger("dials.command_line.find_spots_server")

help_message = """\
A client/server version of dials.find_spots with additional analysis including
estimation of resolution limits. Intended for quick feedback of image quality
during grid scans and data collections.

On the server machine::

  dials.find_spots_server [nproc=8] [port=1234]

On the client machine::

  dials.find_spots_client [host=hostname] [port=1234] [nproc=8] /path/to/image.cbf

The client will return a short xml string indicating the number of spots found
and several estimates of the resolution limit.

e.g.::

  <response>
  <image>/path/to/image_0001.cbf</image>
  <spot_count>352</spot_count>
  <spot_count_no_ice>263</spot_count_no_ice>
  <d_min>1.46</d_min>
  <d_min_method_1>1.92</d_min_method_1>
  <d_min_method_2>1.68</d_min_method_2>
  <total_intensity>56215</total_intensity>
  </response>


* ``spot_count`` is the total number of spots found in given image
* ``spot_count_no_ice`` is the number of spots found excluding those at resolutions
  where ice rings may be found
* ``d_min_method_1`` is equivalent to distl's resolution estimate method 1
* ``d_min_method_2`` is equivalent to distl's resolution estimate method 2
* ``total_intensity`` is the total intensity of all strong spots excluding those
  at resolutions where ice rings may be found

Any valid ``dials.find_spots`` parameter may be passed to
``dials.find_spots_client``, e.g.::

  dials.find_spots_client /path/to/image.cbf min_spot_size=2 d_min=2

To stop the server::

  dials.find_spots_client stop [host=hostname] [port=1234]
"""

stop = False


def _filter_by_resolution(experiments, reflections, d_min=None, d_max=None):
    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)
    d_star_sq = flex.pow2(reflections["rlp"].norms())
    reflections["d"] = uctbx.d_star_sq_as_d(d_star_sq)
    # Filter based on resolution
    if d_min is not None:
        selection = reflections["d"] >= d_min
        reflections = reflections.select(selection)
        logger.debug(f"Selected {len(reflections)} reflections with d >= {d_min:f}")

    # Filter based on resolution
    if d_max is not None:
        selection = reflections["d"] <= d_max
        reflections = reflections.select(selection)
        logger.debug(f"Selected {len(reflections)} reflections with d <= {d_max:f}")
    return reflections


def work(filename, cl=None):
    if cl is None:
        cl = []

    phil_scope = libtbx.phil.parse(
        """\
ice_rings {
  filter = True
    .type = bool
  width = 0.004
    .type = float(value_min=0.0)
}
index = False
  .type = bool
integrate = False
  .type = bool
indexing_min_spots = 10
  .type = int(value_min=1)
"""
    )
    interp = phil_scope.command_line_argument_interpreter()
    params, unhandled = interp.process_and_fetch(
        cl, custom_processor="collect_remaining"
    )
    filter_ice = params.extract().ice_rings.filter
    ice_rings_width = params.extract().ice_rings.width
    index = params.extract().index
    integrate = params.extract().integrate
    indexing_min_spots = params.extract().indexing_min_spots

    interp = find_spots_phil_scope.command_line_argument_interpreter()
    phil_scope, unhandled = interp.process_and_fetch(
        unhandled, custom_processor="collect_remaining"
    )
    logger.info("The following spotfinding parameters have been modified:")
    logger.info(find_spots_phil_scope.fetch_diff(source=phil_scope).as_str())
    params = phil_scope.extract()
    # no need to write the hot mask in the server/client
    params.spotfinder.write_hot_mask = False
    experiments = ExperimentListFactory.from_filenames([filename])
    if params.spotfinder.scan_range and len(experiments) > 1:
        # This means we've imported a sequence of still image: select
        # only the experiment, i.e. image, we're interested in
        ((start, end),) = params.spotfinder.scan_range
        experiments = experiments[start - 1 : end]

    # Avoid overhead of calculating per-pixel resolution masks in spotfinding
    # and instead perform post-filtering of spot centroids by resolution
    d_min = params.spotfinder.filter.d_min
    d_max = params.spotfinder.filter.d_max
    params.spotfinder.filter.d_min = None
    params.spotfinder.filter.d_max = None

    t0 = time.perf_counter()
    reflections = flex.reflection_table.from_observations(experiments, params)

    if d_min or d_max:
        reflections = _filter_by_resolution(
            experiments, reflections, d_min=d_min, d_max=d_max
        )

    t1 = time.perf_counter()
    logger.info("Spotfinding took %.2f seconds", t1 - t0)

    imageset = experiments.imagesets()[0]
    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)
    stats = per_image_analysis.stats_for_reflection_table(
        reflections, filter_ice=filter_ice, ice_rings_width=ice_rings_width
    )._asdict()
    t2 = time.perf_counter()
    logger.info("Resolution analysis took %.2f seconds", t2 - t1)

    if index and stats["n_spots_no_ice"] > indexing_min_spots:
        logging.basicConfig(stream=sys.stdout, level=logging.INFO)

        interp = index_phil_scope.command_line_argument_interpreter()
        phil_scope, unhandled = interp.process_and_fetch(
            unhandled, custom_processor="collect_remaining"
        )
        logger.info("The following indexing parameters have been modified:")
        index_phil_scope.fetch_diff(source=phil_scope).show()
        params = phil_scope.extract()

        if (
            imageset.get_goniometer() is not None
            and imageset.get_scan() is not None
            and imageset.get_scan().is_still()
        ):
            imageset.set_goniometer(None)
            imageset.set_scan(None)

        try:
            idxr = indexer.Indexer.from_parameters(
                reflections, experiments, params=params
            )
            indexing_results = []
            idxr.index()
            indexed_sel = idxr.refined_reflections.get_flags(
                idxr.refined_reflections.flags.indexed
            )
            indexed_sel &= ~(
                idxr.refined_reflections.get_flags(
                    idxr.refined_reflections.flags.centroid_outlier
                )
            )
            for i_expt, expt in enumerate(idxr.refined_experiments):
                sel = idxr.refined_reflections["id"] == i_expt
                sel &= indexed_sel
                indexing_results.append(
                    {
                        "crystal": expt.crystal.to_dict(),
                        "n_indexed": sel.count(True),
                        "fraction_indexed": sel.count(True) / sel.size(),
                    }
                )
            stats["lattices"] = indexing_results
            stats["n_indexed"] = indexed_sel.count(True)
            stats["fraction_indexed"] = indexed_sel.count(True) / len(reflections)
        except Exception as e:
            logger.error(e)
            stats["error"] = str(e)
        finally:
            t3 = time.perf_counter()
            logger.info("Indexing took %.2f seconds", t3 - t2)

        if integrate and "lattices" in stats:
            interp = integrate_phil_scope.command_line_argument_interpreter()
            phil_scope, unhandled = interp.process_and_fetch(
                unhandled, custom_processor="collect_remaining"
            )
            logger.error("The following integration parameters have been modified:")
            integrate_phil_scope.fetch_diff(source=phil_scope).show()
            params = phil_scope.extract()

            try:
                params.profile.gaussian_rs.min_spots = 0

                experiments = idxr.refined_experiments
                reference = idxr.refined_reflections

                predicted = flex.reflection_table.from_predictions_multi(
                    experiments,
                    dmin=params.prediction.d_min,
                    dmax=params.prediction.d_max,
                    margin=params.prediction.margin,
                    force_static=params.prediction.force_static,
                )

                matched, reference, unmatched = predicted.match_with_reference(
                    reference
                )
                assert len(matched) == len(predicted)
                assert matched.count(True) <= len(reference)
                if matched.count(True) == 0:
                    raise Sorry(
                        """
            Invalid input for reference reflections.
            Zero reference spots were matched to predictions
          """
                    )
                elif matched.count(True) != len(reference):
                    logger.info("")
                    logger.info("*" * 80)
                    logger.info(
                        "Warning: %d reference spots were not matched to predictions",
                        len(reference) - matched.count(True),
                    )
                    logger.info("*" * 80)
                    logger.info("")

                # Compute the profile model
                experiments = ProfileModelFactory.create(params, experiments, reference)

                # Compute the bounding box
                predicted.compute_bbox(experiments)

                # Create the integrator
                integrator = create_integrator(params, experiments, predicted)

                # Integrate the reflections
                reflections = integrator.integrate()

                # print len(reflections)

                stats["integrated_intensity"] = flex.sum(
                    reflections["intensity.sum.value"]
                )
            except Exception as e:
                logger.error(e)
                stats["error"] = str(e)
            finally:
                t4 = time.perf_counter()
                logger.info("Integration took %.2f seconds", t4 - t3)

    return stats


class handler(server_base.BaseHTTPRequestHandler):
    def do_GET(self):
        """Respond to a GET request."""
        if self.path == "/Ctrl-C":
            self.send_response(200)
            self.end_headers()

            global stop
            stop = True
            return

        filename = self.path.split(";")[0]
        params = self.path.split(";")[1:]

        # If we're passing a url through, then unquote and ignore leading /
        if "%3A//" in filename:
            filename = urllib.parse.unquote(filename[1:])

        d = {"image": filename}

        try:
            stats = work(filename, params)
            d.update(stats)
            response = 200
        except Exception as e:
            d["error"] = str(e)
            response = 500

        self.send_response(response)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        response = json.dumps(d).encode()
        self.wfile.write(response)


def serve(httpd):
    try:
        while not stop:
            httpd.handle_request()
    except KeyboardInterrupt:
        pass


phil_scope = libtbx.phil.parse(
    """\
nproc = Auto
  .type = int(value_min=1)
port = 1701
  .type = int(value_min=1)
"""
)


def main(nproc, port):
    server_class = server_base.HTTPServer
    httpd = server_class(("", port), handler)
    print(time.asctime(), "Serving %d processes on port %d" % (nproc, port))

    for j in range(nproc - 1):
        proc = multiprocessing.Process(target=serve, args=(httpd,))
        proc.daemon = True
        proc.start()
    serve(httpd)
    httpd.server_close()
    print(time.asctime(), "done")


@show_mail_handle_errors()
def run(args=None):
    usage = "dials.find_spots_server [options]"

    # Python 3.8 on macOS... needs fork
    if sys.hexversion >= 0x3080000 and sys.platform == "darwin":
        multiprocessing.set_start_method("fork")

    parser = ArgumentParser(usage=usage, phil=phil_scope, epilog=help_message)
    params, options = parser.parse_args(args, show_diff_phil=True)
    if params.nproc is libtbx.Auto:
        params.nproc = number_of_processors(return_value_if_unknown=-1)
    main(params.nproc, params.port)


if __name__ == "__main__":
    run()



from __future__ import annotations


def profile2d(p, vmin=None, vmax=None):
    import string

    from dials.array_family import flex

    if vmin is None:
        vmin = flex.min(p)
    if vmax is None:
        vmax = flex.max(p)
    assert vmax >= vmin
    dv = vmax - vmin
    if dv == 0:
        c = 0
        m = 0
    else:
        m = 35.0 / dv
        c = -m * vmin
    lookup = string.digits + string.ascii_uppercase
    ny, nx = p.all()
    text = ""
    for j in range(ny):
        for i in range(nx):
            v = int(m * p[j, i] + c)
            if v < 0:
                v = 0
            elif v > 35:
                v = 35
            t = lookup[v]
            text += t + " "
        text += "\n"
    return text


def profile3d(p, vmin=None, vmax=None):
    """Print a 3D profile."""
    from dials.array_family import flex

    if vmin is None:
        vmin = flex.min(p)
    if vmax is None:
        vmax = flex.max(p)
    nz, ny, nx = p.all()
    text = []
    for k in range(nz):
        p2 = p[k : k + 1, :, :]
        p2.reshape(flex.grid(ny, nx))
        text.append(profile2d(p2, vmin=vmin, vmax=vmax))
    return "\n".join(text)


if __name__ == "__main__":

    from dials.array_family import flex

    a1 = flex.double(
        [
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )
    a2 = flex.double(
        [
            [0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 1, 2, 1, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )
    a3 = flex.double(
        [
            [0, 0, 1, 0, 0],
            [0, 1, 2, 1, 0],
            [1, 2, 5, 2, 1],
            [0, 1, 2, 1, 0],
            [0, 0, 1, 0, 0],
        ]
    )
    a4 = flex.double(
        [
            [0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 1, 2, 1, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )
    a5 = flex.double(
        [
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ]
    )
    a1.reshape(flex.grid(1, 5, 5))
    a2.reshape(flex.grid(1, 5, 5))
    a3.reshape(flex.grid(1, 5, 5))
    a4.reshape(flex.grid(1, 5, 5))
    a5.reshape(flex.grid(1, 5, 5))
    a = flex.double(flex.grid(5, 5, 5))
    a[0:1, :, :] = a1
    a[1:2, :, :] = a2
    a[2:3, :, :] = a3
    a[3:4, :, :] = a4
    a[4:5, :, :] = a5
    a = a * 1000

    print(profile3d(a))



from __future__ import annotations

import itertools
import logging
import math
from io import StringIO

import pkg_resources

import libtbx.phil
import scitbx.matrix
from dxtbx.model.experiment_list import Experiment, ExperimentList
from scitbx.array_family import flex

from dials.algorithms.indexing import indexer
from dials.algorithms.indexing.basis_vector_search import combinations, optimise

from .low_res_spot_match import LowResSpotMatch
from .strategy import Strategy

__all__ = ["Strategy", "LowResSpotMatch"]


logger = logging.getLogger(__name__)


basis_vector_search_phil_str = """\
basis_vector_combinations
    .expert_level = 1
{
    max_combinations = None
        .type = int(value_min=1)
        .help = "Maximum number of basis vector combinations to test for agreement"
                "with input symmetry."
    max_refine = Auto
        .type = int(value_min=1)
        .help = "Maximum number of putative crystal models to test. Default"
                "for rotation sequences: 50, for still images: 5"
        .expert_level = 1
    sys_absent_threshold = 0.9
        .type = float(value_min=0.0, value_max=1.0)
    solution_scorer = filter *weighted
        .type = choice
        .expert_level = 1
    filter
        .expert_level = 1
    {
        check_doubled_cell = True
            .type = bool
        likelihood_cutoff = 0.8
            .type = float(value_min=0, value_max=1)
        volume_cutoff = 1.25
            .type = float(value_min=1)
        n_indexed_cutoff = 0.9
            .type = float(value_min=0, value_max=1)
    }
    weighted
        .expert_level = 1
    {
        power = 1
            .type = int(value_min=1)
        volume_weight = 1
            .type = float(value_min=0)
        n_indexed_weight = 1
            .type = float(value_min=0)
        rmsd_weight = 1
            .type = float(value_min=0)
    }
}

method = None
    .type = choice

optimise_initial_basis_vectors = False
    .type = bool
    .expert_level = 2
"""

basis_vector_search_phil_scope = libtbx.phil.parse(basis_vector_search_phil_str)

methods = []
for entry_point in itertools.chain(
    pkg_resources.iter_entry_points("dials.index.basis_vector_search"),
    pkg_resources.iter_entry_points("dials.index.lattice_search"),
):
    ext_master_scope = libtbx.phil.parse(
        """
%s
.expert_level=1
.help=%s
{}
        """
        % (entry_point.name, entry_point.load().phil_help)
    )
    ext_phil_scope = ext_master_scope.get_without_substitution(entry_point.name)
    assert len(ext_phil_scope) == 1
    ext_phil_scope = ext_phil_scope[0]
    ext_phil_scope.adopt_scope(entry_point.load().phil_scope)
    basis_vector_search_phil_scope.adopt_scope(ext_master_scope)
    methods.append(entry_point.name)
basis_vector_search_phil_scope.adopt_scope(
    libtbx.phil.parse(
        "method = "
        + " ".join("*" + m if m == "fft3d" else m for m in methods)
        + "\n    .type = choice"
    )
)


class LatticeSearch(indexer.Indexer):
    def __init__(self, reflections, experiments, params):
        super().__init__(reflections, experiments, params)

        self._lattice_search_strategy = None
        for entry_point in pkg_resources.iter_entry_points(
            "dials.index.lattice_search"
        ):
            if entry_point.name == self.params.method:
                strategy_class = entry_point.load()
                self._lattice_search_strategy = strategy_class(
                    target_symmetry_primitive=self._symmetry_handler.target_symmetry_primitive,
                    max_lattices=self.params.basis_vector_combinations.max_refine,
                    params=getattr(self.params, entry_point.name, None),
                )
                break

    def find_candidate_crystal_models(self):

        candidate_crystal_models = []
        if self._lattice_search_strategy:
            candidate_crystal_models = (
                self._lattice_search_strategy.find_crystal_models(
                    self.reflections, self.experiments
                )
            )
        return candidate_crystal_models

    def find_lattices(self):
        self.candidate_crystal_models = self.find_candidate_crystal_models()

        crystal_model, n_indexed = self.choose_best_orientation_matrix(
            self.candidate_crystal_models
        )
        if crystal_model is not None:
            crystal_models = [crystal_model]
        else:
            crystal_models = []
        experiments = ExperimentList()
        for cm in crystal_models:
            for expt in self.experiments:
                experiments.append(
                    Experiment(
                        imageset=expt.imageset,
                        beam=expt.beam,
                        detector=expt.detector,
                        goniometer=expt.goniometer,
                        scan=expt.scan,
                        crystal=cm,
                    )
                )
        return experiments

    def choose_best_orientation_matrix(self, candidate_orientation_matrices):

        from dials.algorithms.indexing import model_evaluation

        solution_scorer = self.params.basis_vector_combinations.solution_scorer
        if solution_scorer == "weighted":
            weighted_params = self.params.basis_vector_combinations.weighted
            solutions = model_evaluation.ModelRankWeighted(
                power=weighted_params.power,
                volume_weight=weighted_params.volume_weight,
                n_indexed_weight=weighted_params.n_indexed_weight,
                rmsd_weight=weighted_params.rmsd_weight,
            )
        else:
            filter_params = self.params.basis_vector_combinations.filter
            solutions = model_evaluation.ModelRankFilter(
                check_doubled_cell=filter_params.check_doubled_cell,
                likelihood_cutoff=filter_params.likelihood_cutoff,
                volume_cutoff=filter_params.volume_cutoff,
                n_indexed_cutoff=filter_params.n_indexed_cutoff,
            )

        args = []

        for cm in candidate_orientation_matrices:
            sel = self.reflections["id"] == -1
            if self.d_min is not None:
                sel &= 1 / self.reflections["rlp"].norms() > self.d_min
            xo, yo, zo = self.reflections["xyzobs.mm.value"].parts()
            imageset_id = self.reflections["imageset_id"]
            experiments = ExperimentList()
            for i_expt, expt in enumerate(self.experiments):
                # XXX Not sure if we still need this loop over self.experiments
                if expt.scan is not None:
                    start, end = expt.scan.get_oscillation_range()
                    if (end - start) > 360:
                        # only use reflections from the first 360 degrees of the scan
                        sel.set_selected(
                            (imageset_id == i_expt)
                            & (zo > ((start * math.pi / 180) + 2 * math.pi)),
                            False,
                        )
                experiments.append(
                    Experiment(
                        imageset=expt.imageset,
                        beam=expt.beam,
                        detector=expt.detector,
                        goniometer=expt.goniometer,
                        scan=expt.scan,
                        crystal=cm,
                    )
                )
            refl = self.reflections.select(sel)
            self.index_reflections(experiments, refl)
            if refl.get_flags(refl.flags.indexed).count(True) == 0:
                continue

            from rstbx.dps_core.cell_assessment import SmallUnitCellVolume

            from dials.algorithms.indexing import non_primitive_basis

            threshold = self.params.basis_vector_combinations.sys_absent_threshold
            if threshold and (
                self._symmetry_handler.target_symmetry_primitive is None
                or self._symmetry_handler.target_symmetry_primitive.unit_cell() is None
            ):
                try:
                    non_primitive_basis.correct(
                        experiments, refl, self._assign_indices, threshold
                    )
                    if refl.get_flags(refl.flags.indexed).count(True) == 0:
                        continue
                except SmallUnitCellVolume:
                    logger.debug(
                        "correct_non_primitive_basis SmallUnitCellVolume error for unit cell %s:",
                        experiments[0].crystal.get_unit_cell(),
                    )
                    continue
                except RuntimeError as e:
                    if "Krivy-Gruber iteration limit exceeded" in str(e):
                        logger.debug(
                            "correct_non_primitive_basis Krivy-Gruber iteration limit exceeded error for unit cell %s:",
                            experiments[0].crystal.get_unit_cell(),
                        )
                        continue
                    raise
                if (
                    experiments[0].crystal.get_unit_cell().volume()
                    < self.params.min_cell_volume
                ):
                    continue

            if self.params.known_symmetry.space_group is not None:
                new_crystal, _ = self._symmetry_handler.apply_symmetry(
                    experiments[0].crystal
                )
                if new_crystal is None:
                    continue
                experiments[0].crystal.update(new_crystal)

            args.append((experiments, refl))
            if len(args) == self.params.basis_vector_combinations.max_refine:
                break

        from libtbx import easy_mp

        evaluator = model_evaluation.ModelEvaluation(self.all_params)
        results = easy_mp.parallel_map(
            evaluator.evaluate,
            args,
            iterable_type=easy_mp.posiargs,
            processes=self.params.nproc,
            preserve_exception_message=True,
        )

        for soln in results:
            if soln is None:
                continue
            solutions.append(soln)

        if len(solutions):
            logger.info("Candidate solutions:")
            logger.info(str(solutions))
            best_model = solutions.best_model()
            logger.debug("best model_likelihood: %.2f", best_model.model_likelihood)
            logger.debug("best n_indexed: %i", best_model.n_indexed)
            self.hkl_offset = best_model.hkl_offset
            return best_model.crystal, best_model.n_indexed
        else:
            return None, None


class BasisVectorSearch(LatticeSearch):
    def __init__(self, reflections, experiments, params):
        super().__init__(reflections, experiments, params)

        strategy_class = None
        for entry_point in pkg_resources.iter_entry_points(
            "dials.index.basis_vector_search"
        ):
            if entry_point.name == params.indexing.method:
                strategy_class = entry_point.load()
                break
        if not strategy_class:
            raise RuntimeError(
                f"Unknown basis vector search strategy: {params.indexing.method}"
            )

        target_unit_cell = None
        if (
            self._symmetry_handler.target_symmetry_primitive is not None
            and self._symmetry_handler.target_symmetry_primitive.unit_cell() is not None
        ):
            target_unit_cell = (
                self._symmetry_handler.target_symmetry_primitive.unit_cell()
            )
        self._basis_vector_search_strategy = strategy_class(
            max_cell=self.params.max_cell,
            min_cell=self.params.min_cell,
            target_unit_cell=target_unit_cell,
            params=getattr(self.params, entry_point.name),
        )

    def find_candidate_basis_vectors(self):
        self.d_min = self.params.refinement_protocol.d_min_start
        sel = self.reflections["id"] == -1
        if self.d_min is not None:
            sel &= 1 / self.reflections["rlp"].norms() > self.d_min
        reflections = self.reflections.select(sel)
        (
            self.candidate_basis_vectors,
            used_in_indexing,
        ) = self._basis_vector_search_strategy.find_basis_vectors(reflections["rlp"])
        self._used_in_indexing = sel.iselection().select(used_in_indexing)
        if self.d_min is None:
            self.d_min = flex.min(
                1 / self.reflections["rlp"].select(self._used_in_indexing).norms()
            )

        self.debug_show_candidate_basis_vectors()
        return self.candidate_basis_vectors

    def find_candidate_crystal_models(self):
        self.find_candidate_basis_vectors()
        if self.params.optimise_initial_basis_vectors:
            self.optimise_basis_vectors()
        candidate_crystal_models = self.find_candidate_orientation_matrices(
            self.candidate_basis_vectors
        )
        return candidate_crystal_models

    def find_candidate_orientation_matrices(self, candidate_basis_vectors):

        candidate_crystal_models = combinations.candidate_orientation_matrices(
            candidate_basis_vectors,
            max_combinations=self.params.basis_vector_combinations.max_combinations,
        )
        if self._symmetry_handler.target_symmetry_reference_setting is not None:
            target_symmetry = self._symmetry_handler.target_symmetry_reference_setting
        elif self._symmetry_handler.target_symmetry_primitive is not None:
            target_symmetry = self._symmetry_handler.target_symmetry_primitive
        else:
            target_symmetry = None
        if target_symmetry is not None:
            candidate_crystal_models = combinations.filter_known_symmetry(
                candidate_crystal_models,
                target_symmetry,
                relative_length_tolerance=self.params.known_symmetry.relative_length_tolerance,
                absolute_angle_tolerance=self.params.known_symmetry.absolute_angle_tolerance,
                max_delta=self.params.known_symmetry.max_delta,
            )

        if self.refined_experiments is not None and len(self.refined_experiments) > 0:
            candidate_crystal_models = combinations.filter_similar_orientations(
                candidate_crystal_models,
                self.refined_experiments.crystals(),
                minimum_angular_separation=self.params.multiple_lattice_search.minimum_angular_separation,
            )

        return candidate_crystal_models

    def optimise_basis_vectors(self):

        optimised_basis_vectors = optimise.optimise_basis_vectors(
            self.reflections["rlp"].select(self._used_in_indexing),
            self.candidate_basis_vectors,
        )
        self.candidate_basis_vectors = [
            scitbx.matrix.col(v) for v in optimised_basis_vectors
        ]

    def debug_show_candidate_basis_vectors(self):

        vectors = self.candidate_basis_vectors

        logger.debug("Candidate basis vectors:")
        for i, v in enumerate(vectors):
            logger.debug(f"{i} {v.length()}")  # , vector_heights[i]

        if self.params.debug:
            # print a table of the angles between each pair of vectors
            s = StringIO()

            angles = flex.double(len(vectors) ** 2)
            angles.reshape(flex.grid(len(vectors), len(vectors)))

            for i in range(len(vectors)):
                v_i = vectors[i]
                for j in range(i + 1, len(vectors)):
                    v_j = vectors[j]
                    angles[i, j] = v_i.angle(v_j, deg=True)

            print((" " * 7), end=" ", file=s)
            for i in range(len(vectors)):
                print(f"{vectors[i].length():7.3f}", end=" ", file=s)
            print(file=s)
            for i in range(len(vectors)):
                print(f"{vectors[i].length():7.3f}", end=" ", file=s)
                for j in range(len(vectors)):
                    if j <= i:
                        print((" " * 7), end=" ", file=s)
                    else:
                        print(f"{angles[i, j]:5.1f}  ", end=" ", file=s)
                print(file=s)

            logger.debug(s.getvalue())



from __future__ import annotations

import pytest

from cctbx import sgtbx
from dxtbx.serialize import load

from dials.algorithms.merging import french_wilson, merge
from dials.array_family import flex

MAX_SIGF_OVER_F_ACENTRIC = max(
    sigf / f for f, sigf in zip(french_wilson.ac_zf, french_wilson.ac_zf_sd)
)
MAX_SIGF_OVER_F_CENTRIC = max(
    sigf / f for f, sigf in zip(french_wilson.c_zf, french_wilson.c_zf_sd)
)


def check_french_wilson_amplitudes(amplitudes):
    assert amplitudes.is_xray_amplitude_array()
    assert flex.min(amplitudes.data()) >= 0
    assert flex.min(amplitudes.sigmas()) >= 0
    sigF_over_F = amplitudes.sigmas() / amplitudes.data()
    is_centric = amplitudes.centric_flags().data()
    assert flex.max(sigF_over_F.select(~is_centric)) <= MAX_SIGF_OVER_F_ACENTRIC
    if is_centric.count(True):
        assert flex.max(sigF_over_F.select(is_centric)) <= MAX_SIGF_OVER_F_CENTRIC


def test_french_wilson_insulin(dials_data):
    insulin = dials_data("insulin_processed", pathlib=True)
    expts = load.experiment_list(insulin / "scaled.expt", check_format=False)
    refls = flex.reflection_table.from_file(insulin / "scaled.refl")
    merged, _, _ = merge.merge(expts, refls)
    merged_intensities = merged.array()

    amplitudes = french_wilson.french_wilson(merged_intensities)
    check_french_wilson_amplitudes(amplitudes)


@pytest.mark.parametrize("space_group_symbol", [None, "P1"])
def test_french_wilson_l_cysteine(dials_data, space_group_symbol):
    l_cysteine = dials_data("l_cysteine_4_sweeps_scaled")
    expts = load.experiment_list(l_cysteine / "scaled_30.expt", check_format=False)
    refls = flex.reflection_table.from_file(l_cysteine / "scaled_30.refl")
    if space_group_symbol:
        expts[0].crystal.set_space_group(
            sgtbx.space_group_info(space_group_symbol).group()
        )
    merged, _, _ = merge.merge(expts, refls)
    merged_intensities = merged.array()

    amplitudes = french_wilson.french_wilson(merged_intensities)
    check_french_wilson_amplitudes(amplitudes)



"""
Tests for the reflection selection algorithm.
"""

from __future__ import annotations

import itertools
from unittest.mock import Mock

from cctbx import sgtbx, uctbx
from libtbx import phil
from scitbx import sparse

from dials.algorithms.scaling.Ih_table import IhTable
from dials.algorithms.scaling.reflection_selection import (
    _loop_over_class_matrix,
    calculate_scaling_subset_ranges,
    calculate_scaling_subset_ranges_with_E2,
    select_connected_reflections_across_datasets,
)
from dials.array_family import flex


def test_select_connected_reflections_across_datasets():
    """Test the basic cross-dataset reflection selection algorithm.

    Make three reflection tables with the following reflections:
               symmetry groups
               0  1  2  3  4  5  6
            0  3  3  2  0  1  1  1
    classes 1  0  2  0  0  3  2  1
            2  2  1  1  5  0  4  0

    With target=5, expect:
    number of chosen reflections per class: [8, 7, 7]
    symmetry groups used:                   [1, 5, 0, 4]
    """

    n1 = [3, 3, 2, 0, 1, 1, 1]
    n2 = [0, 2, 0, 0, 3, 2, 1]
    n3 = [2, 1, 1, 5, 0, 4, 0]

    def make_refl_table(n_list, class_idx=0):
        """Make a reflection table with groups based on n_list."""
        r1 = flex.reflection_table()
        miller_indices = [[(0, 0, i + 1)] * n for i, n in enumerate(n_list)]
        r1["miller_index"] = flex.miller_index(
            list(itertools.chain.from_iterable(miller_indices))
        )
        r1["intensity"] = flex.double(sum(n_list), 1)
        r1["variance"] = flex.double(sum(n_list), 1)
        r1["inverse_scale_factor"] = flex.double(sum(n_list), 1)
        r1["class_index"] = flex.int(sum(n_list), class_idx)
        return r1

    reflections = [
        make_refl_table(n1, 0),
        make_refl_table(n2, 1),
        make_refl_table(n3, 2),
    ]

    space_group = sgtbx.space_group("P1")
    table = IhTable(reflections, space_group)
    experiment = Mock()
    experiment.crystal.get_space_group.return_value = space_group
    experiment.crystal.get_unit_cell.return_value = uctbx.unit_cell(
        (10, 10, 10, 90, 90, 90)
    )
    indices, datset_ids = select_connected_reflections_across_datasets(
        table, experiment, Isigma_cutoff=0.0, min_total=30, n_resolution_bins=1
    )
    assert list(indices) == [0, 1, 2, 3, 4, 5, 8, 9] + [0, 1, 2, 3, 4, 5, 6] + [
        0,
        1,
        2,
        9,
        10,
        11,
        12,
    ]
    assert list(datset_ids) == [0] * 8 + [1] * 7 + [2] * 7

    # now test again with a higher min_total
    indices, datset_ids = select_connected_reflections_across_datasets(
        table, experiment, Isigma_cutoff=0.0, min_total=10 * 3 * 4, n_resolution_bins=1
    )
    assert list(datset_ids) == [0] * 11 + [1] * 8 + [2] * 13


def test_loop_over_class_matrix():
    """Test a few different limits of the method.

    { 3, 1, 3, 2, 1, 1, 0 },
    { 2, 2, 0, 0, 3, 1, 0 },
    { 1, 4, 2, 1, 0, 0, 5 },
    """
    sorted_class_matrix = sparse.matrix(
        3,
        7,
        elements_by_columns=[
            {0: 3, 1: 2, 2: 1},
            {0: 1, 1: 2, 2: 4},
            {0: 3, 2: 2},
            {0: 2, 2: 1},
            {0: 1, 1: 3},
            {0: 1},
            {2: 5},
        ],
    )

    # first test if don't meet the minimum number
    total_in_classes, cols_not_used = _loop_over_class_matrix(
        sorted_class_matrix, 1, 8, 20
    )
    assert list(cols_not_used) == [2, 3, 4, 5, 6]
    assert list(total_in_classes) == [4.0, 4.0, 5.0]
    # now test if max per bin reached
    total_in_classes, cols_not_used = _loop_over_class_matrix(
        sorted_class_matrix, 2, 3, 8
    )
    assert list(cols_not_used) == [2, 3, 4, 5, 6]
    assert list(total_in_classes) == [4.0, 4.0, 5.0]

    # now test if request more than available - should find all
    total_in_classes, cols_not_used = _loop_over_class_matrix(
        sorted_class_matrix, 9, 50, 100
    )
    assert not cols_not_used
    assert list(total_in_classes) == [11.0, 7.0, 13.0]


def generated_param():
    """Generate a param phil scope."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
  """,
        process_includes=True,
    )
    return phil_scope.extract()


def generated_refl_for_subset_calculation():
    """Create a reflection table suitable for splitting into blocks."""
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    reflections["variance"] = flex.double(6, 1.0)
    reflections["d"] = flex.double([0.8, 2.1, 2.0, 1.4, 1.6, 2.5])
    reflections["partiality"] = flex.double(6, 1.0)
    reflections.set_flags(flex.bool(6, True), reflections.flags.integrated)
    reflections.set_flags(flex.bool(6, False), reflections.flags.bad_for_scaling)
    return reflections


def test_selection_scaling_subset_ranges_with_E2():
    """Test the scaling subset calculation with E2 range."""
    test_params = generated_param()
    rt = generated_refl_for_subset_calculation()
    rt["Esq"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    test_params.reflection_selection.E2_range = 0.8, 5.0
    test_params.reflection_selection.d_range = 1.0, 5.0  # all but first
    test_params.reflection_selection.Isigma_range = 0.9, 5.5  # all but last
    sel = calculate_scaling_subset_ranges_with_E2(rt, test_params)
    assert list(sel) == [False, True, True, True, True, False]
    rt["Esq"] = flex.double([1.0, 1.0, 1.0, 0.1, 6.0, 1.0])
    sel = calculate_scaling_subset_ranges_with_E2(rt, test_params)
    assert list(sel) == [False, True, True, False, False, False]


def test_selection_scaling_subset_ranges():
    """Test the scaling subset calculation with E2 range."""
    test_params = generated_param()
    rt = generated_refl_for_subset_calculation()
    test_params.reflection_selection.E2_range = 0.8, 5.0
    test_params.reflection_selection.d_range = 1.0, 5.0  # all but first
    test_params.reflection_selection.Isigma_range = 0.9, 5.5  # all but last
    sel = calculate_scaling_subset_ranges(rt, test_params)
    assert list(sel) == [False, True, True, True, True, False]



"""
Test the command line script dials.scale, for successful completion.
"""

from __future__ import annotations

import json

import procrunner
import pytest

import iotbx.merging_statistics
import iotbx.mtz
from cctbx import uctbx
from dxtbx.model import Beam, Crystal, Detector, Experiment, Goniometer, Scan
from dxtbx.model.experiment_list import ExperimentList
from dxtbx.serialize import load
from libtbx import phil

from dials.algorithms.scaling.algorithm import ScalingAlgorithm, prepare_input
from dials.array_family import flex
from dials.command_line import merge, report, scale
from dials.util.options import ArgumentParser


def run_one_scaling(working_directory, argument_list):
    """Run the dials.scale algorithm."""
    command = ["dials.scale"] + argument_list
    print(command)
    result = procrunner.run(command, working_directory=working_directory)
    print(result.stderr)
    assert not result.returncode and not result.stderr
    assert (working_directory / "scaled.expt").is_file()
    assert (working_directory / "scaled.refl").is_file()
    assert (working_directory / "dials.scale.html").is_file()

    table = flex.reflection_table.from_file(working_directory / "scaled.refl")

    assert "inverse_scale_factor" in table
    assert "inverse_scale_factor_variance" in table


def get_merging_stats(
    scaled_unmerged_mtz,
    anomalous=False,
    n_bins=20,
    use_internal_variance=False,
    eliminate_sys_absent=False,
    data_labels=None,
):
    """Return a merging statistics result from an mtz file."""

    i_obs = iotbx.merging_statistics.select_data(
        str(scaled_unmerged_mtz), data_labels=data_labels
    )
    i_obs = i_obs.customized_copy(anomalous_flag=False, info=i_obs.info())
    result = iotbx.merging_statistics.dataset_statistics(
        i_obs=i_obs,
        n_bins=n_bins,
        anomalous=anomalous,
        use_internal_variance=use_internal_variance,
        eliminate_sys_absent=eliminate_sys_absent,
    )
    return result


def generated_exp(n=1):
    """Generate an experiment list with two experiments."""
    experiments = ExperimentList()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    crystal = Crystal.from_dict(exp_dict)
    scan = Scan(image_range=[0, 90], oscillation=[0.0, 1.0])
    beam = Beam(s0=(0.0, 0.0, 1.01))
    goniometer = Goniometer((1.0, 0.0, 0.0))
    detector = Detector()
    experiments.append(
        Experiment(
            beam=beam,
            scan=scan,
            goniometer=goniometer,
            detector=detector,
            crystal=crystal,
        )
    )
    if n > 1:
        for _ in range(n - 1):
            experiments.append(
                Experiment(
                    beam=beam,
                    scan=scan,
                    goniometer=goniometer,
                    detector=detector,
                    crystal=crystal,
                )
            )
    return experiments


def generated_param():
    """Generate the default scaling parameters object."""
    phil_scope = phil.parse(
        """
      include scope dials.command_line.scale.phil_scope
  """,
        process_includes=True,
    )

    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    return parameters


def generate_test_reflections():
    """Generate a small reflection table."""
    reflections = flex.reflection_table()
    reflections["intensity.sum.value"] = flex.double([1.0, 2.0, 3.0, 4.0])
    reflections["intensity.sum.variance"] = flex.double([1.0, 2.0, 3.0, 4.0])
    reflections["miller_index"] = flex.miller_index(
        [(0, 0, 1), (0, 0, 1), (0, 0, 2), (0, 0, 2)]
    )
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [(0, 0, 1), (0, 0, 2), (0, 0, 3), (0, 0, 4)]
    )
    reflections["id"] = flex.int([0, 0, 0, 0])
    return reflections


def generate_test_input(n=1):
    """Generate params, exps and refls."""
    reflections = []
    for _ in range(n):
        reflections.append(generate_test_reflections())
    return generated_param(), generated_exp(n), reflections


def test_scale_script_prepare_input():
    """Test prepare_input method of scaling script."""

    # test the components of the scaling script directly with a test reflection
    # table, experiments list and params.

    params, exp, reflections = generate_test_input()
    # try to pass in unequal number of reflections and experiments
    reflections.append(generate_test_reflections())
    with pytest.raises(ValueError):
        _ = ScalingAlgorithm(params, exp, reflections)

    params, exp, reflections = generate_test_input()
    # Try to use use_datasets when not identifiers set
    params.dataset_selection.use_datasets = [0]
    with pytest.raises(ValueError):
        _ = ScalingAlgorithm(params, exp, reflections)
    # Try to use use_datasets when not identifiers set
    params.dataset_selection.use_datasets = None
    params.dataset_selection.exclude_datasets = [0]
    with pytest.raises(ValueError):
        _ = ScalingAlgorithm(params, exp, reflections)

    # Now make two experiments with identifiers and select on them
    params, exp, reflections = generate_test_input(n=2)
    exp[0].identifier = "0"
    reflections[0].experiment_identifiers()[0] = "0"
    exp[1].identifier = "1"
    reflections[1].experiment_identifiers()[0] = "1"
    list1 = ExperimentList().append(exp[0])
    list2 = ExperimentList().append(exp[1])
    reflections[0].assert_experiment_identifiers_are_consistent(list1)
    reflections[1].assert_experiment_identifiers_are_consistent(list2)
    params.dataset_selection.use_datasets = [0]
    params, exp, script_reflections = prepare_input(params, exp, reflections)

    assert len(script_reflections) == 1

    # Try again, this time excluding
    params, exp, reflections = generate_test_input(n=2)
    exp[0].identifier = "0"
    reflections[0].experiment_identifiers()[0] = "0"
    exp[1].identifier = "1"
    reflections[1].experiment_identifiers()[1] = "1"
    params.dataset_selection.exclude_datasets = [0]
    params, exp, script_reflections = prepare_input(params, exp, reflections)

    assert len(script_reflections) == 1
    assert script_reflections[0] is reflections[1]

    # Try having two unequal space groups
    params, exp, reflections = generate_test_input(n=2)
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " P 1",
    }
    crystal = Crystal.from_dict(exp_dict)
    exp[0].crystal = crystal
    with pytest.raises(ValueError):
        _ = prepare_input(params, exp, reflections)

    # Test cutting data
    params, exp, reflections = generate_test_input(n=1)
    params.cut_data.d_min = 1.5
    params, _, script_reflections = prepare_input(params, exp, reflections)
    r = script_reflections[0]
    assert list(r.get_flags(r.flags.user_excluded_in_scaling)) == [
        False,
        False,
        True,
        True,
    ]

    # Ensure that the user_excluded_in_scaling flags are reset before applying any new
    # cutoffs by re-passing script_reflections to prepare_input
    params.cut_data.d_min = None
    params, _, script_reflections = prepare_input(params, exp, script_reflections)
    r = script_reflections[0]
    assert list(r.get_flags(r.flags.user_excluded_in_scaling)) == [
        False,
        False,
        False,
        False,
    ]

    params.cut_data.d_max = 1.25
    params, _, script_reflections = prepare_input(params, exp, reflections)
    r = script_reflections[0]
    assert list(r.get_flags(r.flags.user_excluded_in_scaling)) == [
        True,
        True,
        False,
        False,
    ]

    params, exp, reflections = generate_test_input(n=1)
    reflections[0]["partiality"] = flex.double([0.5, 0.8, 1.0, 1.0])
    params.cut_data.partiality_cutoff = 0.75
    _, __, script_reflections = prepare_input(params, exp, reflections)
    r = script_reflections[0]
    assert list(r.get_flags(r.flags.user_excluded_in_scaling)) == [
        True,
        False,
        False,
        False,
    ]


def test_targeted_scaling_against_mtz(dials_data, tmp_path):
    """Test targeted scaling against an mtz generated with dials.scale."""
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refl = location / "scaled_35.refl"
    expt = location / "scaled_35.expt"
    command = ["dials.scale", refl, expt, "unmerged_mtz=unmerged.mtz"]

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "unmerged.mtz").is_file()

    refl = location / "scaled_30.refl"
    expt = location / "scaled_30.expt"
    reference = tmp_path / "unmerged.mtz"
    command = [
        "dials.scale",
        refl,
        expt,
        f"reference={reference}",
        "unmerged_mtz=unmerged_2.mtz",
    ]

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "unmerged_2.mtz").is_file()
    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    assert len(expts) == 1
    result = get_merging_stats(tmp_path / "unmerged_2.mtz")
    assert result.overall.r_pim < 0.024  # at 30/03/22, value was 0.022
    assert result.overall.cc_one_half > 0.995  # at 30/03/22, value was 0.998
    assert result.overall.n_obs > 3150  # at 30/03/22, value was 3188


@pytest.mark.parametrize(
    "option",
    [
        None,
        "reflection_selection.method=random",
        "reflection_selection.method=intensity_ranges",
        "reflection_selection.method=use_all",
        "intensity_choice=sum",
        "intensity=profile",
    ],
)
def test_scale_single_dataset_with_options(dials_data, tmp_path, option):
    """Test different non-default command-line options with a single dataset."""
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    args = [refl_1, expt_1]
    if option:
        args.append(option)
    run_one_scaling(tmp_path, args)


@pytest.fixture(scope="session")
def vmxi_protk_reindexed(dials_data, tmp_path_factory):
    """Reindex the protk data to be in the correct space group."""
    location = dials_data("vmxi_proteinase_k_sweeps", pathlib=True)

    command = [
        "dials.reindex",
        location / "experiments_0.json",
        location / "reflections_0.pickle",
        "space_group=P422",
    ]
    tmp_path = tmp_path_factory.mktemp("vmxi_protk_reindexed")
    procrunner.run(command, working_directory=tmp_path)
    return tmp_path / "reindexed.expt", tmp_path / "reindexed.refl"


@pytest.mark.parametrize(
    ("options", "expected", "tolerances"),
    [
        (["error_model=None"], None, None),
        (
            ["error_model=basic", "basic.minimisation=individual"],
            (0.61, 0.049),
            (0.05, 0.005),
        ),
        (["error_model.basic.a=0.61"], (0.61, 0.049), (1e-6, 0.005)),
        (["error_model.basic.b=0.049"], (0.61, 0.049), (0.05, 1e-6)),
        (
            ["error_model.basic.b=0.02", "error_model.basic.a=1.5"],
            (1.50, 0.02),
            (1e-6, 1e-6),
        ),
        (
            ["error_model=basic", "basic.minimisation=regression"],
            (0.995, 0.051),
            (0.05, 0.005),
        ),
        (
            ["error_model.basic.a=0.99", "basic.minimisation=regression"],
            (0.99, 0.051),
            (1e-6, 0.005),
        ),
        (
            ["error_model.basic.b=0.051", "basic.minimisation=regression"],
            (1.0, 0.051),
            (0.05, 1e-6),
        ),
    ],
)
def test_error_model_options(
    vmxi_protk_reindexed, tmp_path, options, expected, tolerances
):
    """Test different non-default command-line options with a single dataset.

    Current values taken at 14.11.19"""
    expt_1, refl_1 = vmxi_protk_reindexed
    args = [refl_1, expt_1] + list(options)
    run_one_scaling(tmp_path, args)
    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    config = expts[0].scaling_model.configdict
    if not expected:
        assert "error_model_parameters" not in config
    else:
        params = expts[0].scaling_model.configdict["error_model_parameters"]
        print(list(params))
        assert params[0] == pytest.approx(expected[0], abs=tolerances[0])
        assert params[1] == pytest.approx(expected[1], abs=tolerances[1])


@pytest.mark.parametrize(
    "option",
    [
        None,
        "combine.joint_analysis=False",
        "reflection_selection.method=quasi_random",
        "reflection_selection.method=random",
        "reflection_selection.method=intensity_ranges",
        "reflection_selection.method=use_all",
        "anomalous=True",
    ],
)
def test_scale_multiple_datasets_with_options(dials_data, tmp_path, option):
    """Test different non-defaul command-line options with multiple datasets."""
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    refl_2 = data_dir / "25_integrated.pickle"
    expt_2 = data_dir / "25_integrated_experiments.json"
    args = [refl_1, expt_1, refl_2, expt_2]
    if option:
        args.append(option)
    run_one_scaling(tmp_path, args)


def test_scale_physical(dials_data, tmp_path):
    """Test standard scaling of one dataset."""

    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    extra_args = [
        "model=physical",
        "merged_mtz=merged.mtz",
        "error_model=None",
        "intensity_choice=profile",
        "unmerged_mtz=unmerged.mtz",
        "crystal_name=foo",
        "project_name=bar",
        "use_free_set=1",
        "outlier_rejection=simple",
        "json=scaling.json",
    ]
    run_one_scaling(tmp_path, [refl_1, expt_1] + extra_args)
    unmerged_mtz = tmp_path / "unmerged.mtz"
    merged_mtz = tmp_path / "merged.mtz"
    assert unmerged_mtz.is_file()
    assert merged_mtz.is_file()
    assert (tmp_path / "scaling.json").is_file()
    for f in (unmerged_mtz, merged_mtz):
        mtz_obj = iotbx.mtz.object(str(f))
        assert mtz_obj.crystals()[1].name() == "foo"
        assert mtz_obj.crystals()[1].project_name() == "bar"

    # Now inspect output, check it hasn't changed drastically, or if so verify
    # that the new behaviour is more correct and update test accordingly.
    result = get_merging_stats(unmerged_mtz)
    print(result.overall.r_pim, result.overall.cc_one_half, result.overall.n_obs)
    assert result.overall.r_pim < 0.0255  # at 30/01/19, value was 0.02410
    assert result.overall.cc_one_half > 0.9955  # at 30/01/19, value was 0.9960
    assert result.overall.n_obs > 2300  # at 30/01/19, was 2320

    refls = flex.reflection_table.from_file(tmp_path / "scaled.refl")
    n_scaled = refls.get_flags(refls.flags.scaled).count(True)
    assert n_scaled == result.overall.n_obs
    assert n_scaled == refls.get_flags(refls.flags.bad_for_scaling, all=False).count(
        False
    )

    # run again with the concurrent scaling option turned off and the 'standard'
    # outlier rejection
    extra_args = [
        "model=physical",
        "merged_mtz=merged.mtz",
        "unmerged_mtz=unmerged.mtz",
        "use_free_set=1",
        "outlier_rejection=standard",
        "refinement_order=consecutive",
        "intensity_choice=combine",
    ]
    run_one_scaling(tmp_path, [refl_1, expt_1] + extra_args)

    # Now inspect output, check it hasn't changed drastically, or if so verify
    # that the new behaviour is more correct and update test accordingly.
    result = get_merging_stats(tmp_path / "unmerged.mtz")
    assert (
        result.overall.r_pim < 0.024
    )  # at 07/01/19, value was 0.02372, at 30/01/19 was 0.021498
    assert (
        result.overall.cc_one_half > 0.995
    )  # at 07/01/19, value was 0.99568, at 30/01/19 was 0.9961
    assert result.overall.n_obs > 2300  # at 07/01/19, was 2336, at 22/05/19 was 2311


def test_scale_set_absorption_level(dials_data, tmp_path):
    """Test that the absorption parameters are correctly set for the absorption option."""
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    refl = location / "20_integrated.pickle"
    expt = location / "20_integrated_experiments.json"

    # exclude a central region of data to force the failure of the full matrix
    # minimiser due to indeterminate solution of the normal equations. In this
    # case, the error should be caught and scaling can proceed.
    command = [
        "dials.scale",
        refl,
        expt,
        "absorption_level=medium",
        "unmerged_mtz=unmerged.mtz",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    assert expts[0].scaling_model.configdict["lmax"] == 6
    assert expts[0].scaling_model.configdict["abs_surface_weight"] == 5e4
    abs_params = expts[0].scaling_model.components["absorption"].parameters
    result = get_merging_stats(tmp_path / "unmerged.mtz")
    assert result.overall.r_pim < 0.024
    assert result.overall.cc_one_half > 0.995
    assert result.overall.n_obs > 2300

    ## now scale again with different options, but fix the absorption surface to
    # test the correction.fix option.
    command = [
        "dials.scale",
        tmp_path / "scaled.refl",
        tmp_path / "scaled.expt",
        "error_model=None",
        "physical.correction.fix=absorption",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    new_abs_params = expts[0].scaling_model.components["absorption"].parameters
    assert abs_params == new_abs_params


def test_scale_normal_equations_failure(dials_data, tmp_path):
    location = dials_data("l_cysteine_dials_output", pathlib=True)
    refl = location / "20_integrated.pickle"
    expt = location / "20_integrated_experiments.json"

    # exclude a central region of data to force the failure of the full matrix
    # minimiser due to indeterminate solution of the normal equations. In this
    # case, the error should be caught and scaling can proceed.
    command = ["dials.scale", refl, expt, "exclude_images=800:1400"]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()


def test_scale_and_filter_image_group_mode(dials_data, tmp_path):
    """Test the scale and filter command line program."""
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)

    command = [
        "dials.scale",
        "filtering.method=deltacchalf",
        "stdcutoff=4.0",
        "mode=image_group",
        "max_percent_removed=6.0",
        "max_cycles=6",
        "d_min=2.0",
        "group_size=5",
        "unmerged_mtz=unmerged.mtz",
        "scale_and_filter_results=analysis_results.json",
        "error_model=None",
    ]
    for i in [1, 2, 3, 4, 5, 7, 10]:
        command.append(location / f"experiments_{i}.json")
        command.append(location / f"reflections_{i}.pickle")

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "analysis_results.json").is_file()
    result = get_merging_stats(tmp_path / "unmerged.mtz")
    assert result.overall.r_pim < 0.15  # on 04/05/22 on mac, was 0.143,
    assert result.overall.cc_one_half > 0.89  # on 04/05/22 on mac was 0.893
    assert result.overall.n_obs > 20000  # on 04/05/22 on mac was 20188

    with open(tmp_path / "analysis_results.json") as fh:
        analysis_results = json.load(fh)
    assert analysis_results["cycle_results"]["1"]["image_ranges_removed"] == [
        [[16, 24], 4]
    ]
    assert analysis_results["cycle_results"]["2"]["image_ranges_removed"] == []
    assert analysis_results["termination_reason"] == "no_more_removed"


def test_scale_and_filter_termination(dials_data, tmp_path):
    """Test the scale and filter command line program,
    when it terminates with a cycle when going over the max percent removed."""
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)

    command = [
        "dials.scale",
        "filtering.method=deltacchalf",
        "stdcutoff=2.0",
        "max_percent_removed=5.0",
        "max_cycles=8",
        "d_min=2.0",
        "unmerged_mtz=unmerged.mtz",
        "scale_and_filter_results=analysis_results.json",
        "error_model=None",
        "full_matrix=False",
    ]
    for i in [1, 2, 3, 4, 5, 7, 10]:
        command.append(location / f"experiments_{i}.json")
        command.append(location / f"reflections_{i}.pickle")

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "analysis_results.json").is_file()

    with open(tmp_path / "analysis_results.json") as fh:
        analysis_results = json.load(fh)
    assert analysis_results["termination_reason"] == "max_percent_removed"
    assert len(analysis_results["cycle_results"]["1"]["removed_datasets"]) == 1
    refls = flex.reflection_table.from_file(tmp_path / "scaled.refl")
    assert len(set(refls["id"])) == 6
    assert len(set(refls["imageset_id"])) == 6


def test_scale_and_filter_image_group_single_dataset(dials_data, tmp_path):
    """Test the scale and filter deltacchalf.mode=image_group on a
    single data set."""
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    command = [
        "dials.scale",
        data_dir / "20_integrated.pickle",
        data_dir / "20_integrated_experiments.json",
        "filtering.method=deltacchalf",
        "stdcutoff=3.0",
        "mode=image_group",
        "max_cycles=1",
        "scale_and_filter_results=analysis_results.json",
        "error_model=None",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "analysis_results.json").is_file()

    with open(tmp_path / "analysis_results.json") as fh:
        analysis_results = json.load(fh)
    assert analysis_results["cycle_results"]["1"]["image_ranges_removed"] == []
    assert len(analysis_results["cycle_results"].keys()) == 1
    assert analysis_results["termination_reason"] == "no_more_removed"


def test_scale_when_a_dataset_is_filtered_out(dials_data, tmp_path):
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)
    # Modify one of the input tables so that it gets filtered out
    # during scaler initialisation. Triggers the bug referenced in
    # https://github.com/dials/dials/issues/2045
    refls = flex.reflection_table.from_file(location / "reflections_3.pickle")
    refls["partiality"] = flex.double(refls.size(), 0.3)
    refls.as_file(tmp_path / "modified_3.refl")
    command = ["dials.scale", "d_min=2.0", tmp_path / "modified_3.refl"]
    for i in [1, 2, 3, 4]:
        command.append(location / f"experiments_{i}.json")
    for i in [1, 2, 4]:
        command.append(location / f"reflections_{i}.pickle")
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    assert len(expts) == 3


def test_scale_dose_decay_model(dials_data, tmp_path):
    """Test the scale and filter command line program."""
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)
    command = ["dials.scale", "d_min=2.0", "model=dose_decay"]
    for i in [1, 2, 3, 4, 5, 7, 10]:
        command.append(location / f"experiments_{i}.json")
        command.append(location / f"reflections_{i}.pickle")

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "dials.scale.html").is_file()
    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    assert expts[0].scaling_model.id_ == "dose_decay"


def test_scale_best_unit_cell_d_min(dials_data, tmp_path):
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)
    best_unit_cell = uctbx.unit_cell((42, 42, 39, 90, 90, 90))
    d_min = 2
    command = [
        "dials.scale",
        "best_unit_cell=%g,%g,%g,%g,%g,%g" % best_unit_cell.parameters(),
        f"d_min={d_min:g}",
        "unmerged_mtz=unmerged.mtz",
        "merged_mtz=merged.mtz",
    ]
    for i in [1, 2, 3, 4, 5, 7, 10]:
        command.append(location / f"experiments_{i}.json")
        command.append(location / f"reflections_{i}.pickle")
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "unmerged.mtz").is_file()
    assert (tmp_path / "merged.mtz").is_file()
    stats = get_merging_stats(tmp_path / "unmerged.mtz")
    assert stats.overall.d_min >= d_min
    assert stats.crystal_symmetry.unit_cell().parameters() == pytest.approx(
        best_unit_cell.parameters()
    )
    m = iotbx.mtz.object(str(tmp_path / "merged.mtz"))
    for ma in m.as_miller_arrays():
        assert best_unit_cell.parameters() == pytest.approx(ma.unit_cell().parameters())


def test_scale_and_filter_dataset_mode(dials_data, tmp_path):
    """Test the scale and filter command line program."""
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)
    command = [
        "dials.scale",
        "filtering.method=deltacchalf",
        "stdcutoff=1.0",
        "mode=dataset",
        "max_cycles=2",
        "d_min=1.4",
        "output.reflections=filtered.refl",
        "scale_and_filter_results=analysis_results.json",
        "unmerged_mtz=unmerged.mtz",
        "error_model=None",
    ]
    for i in [1, 2, 3, 4, 5, 7, 10]:
        command.append(location / f"experiments_{i}.json")
        command.append(location / f"reflections_{i}.pickle")

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "filtered.refl").is_file()
    assert (tmp_path / "scaled.expt").is_file()
    assert (tmp_path / "analysis_results.json").is_file()

    with open(tmp_path / "analysis_results.json") as fh:
        analysis_results = json.load(fh)
    assert analysis_results["cycle_results"]["1"]["removed_datasets"] == [
        analysis_results["initial_expids_and_image_ranges"][4][0]
    ]
    assert "expids_and_image_ranges" in analysis_results
    refls = flex.reflection_table.from_file(tmp_path / "filtered.refl")
    assert set(refls["id"]) == {0, 1, 2, 3, 5, 6}


def test_scale_array(dials_data, tmp_path):
    """Test a standard dataset - ideally needs a large dataset or full matrix
    round may fail. Currently turning off absorption term to avoid
    overparameterisation and failure of full matrix minimisation."""

    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl = data_dir / "20_integrated.pickle"
    expt = data_dir / "20_integrated_experiments.json"
    extra_args = ["model=array", "array.absorption_correction=0", "full_matrix=0"]

    run_one_scaling(tmp_path, [refl, expt] + extra_args)


def test_multi_scale(dials_data, tmp_path):
    """Test standard scaling of two datasets."""

    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    refl_2 = data_dir / "25_integrated.pickle"
    expt_2 = data_dir / "25_integrated_experiments.json"
    extra_args = [
        "unmerged_mtz=unmerged.mtz",
        "error_model=None",
        "intensity_choice=profile",
        "outlier_rejection=simple",
    ]

    run_one_scaling(tmp_path, [refl_1, refl_2, expt_1, expt_2] + extra_args)

    # Now inspect output, check it hasn't changed drastically, or if so verify
    # that the new behaviour is more correct and update test accordingly.
    result = get_merging_stats(tmp_path / "unmerged.mtz")
    expected_nobs = 5526  # 19/06/20
    assert abs(result.overall.n_obs - expected_nobs) < 30
    assert result.overall.r_pim < 0.0221  # at 22/10/18, value was 0.22037
    assert result.overall.cc_one_half > 0.9975  # at 07/08/18, value was 0.99810
    print(result.overall.r_pim)
    print(result.overall.cc_one_half)

    refls = flex.reflection_table.from_file(tmp_path / "scaled.refl")
    n_scaled = refls.get_flags(refls.flags.scaled).count(True)
    assert n_scaled == result.overall.n_obs
    assert n_scaled == refls.get_flags(refls.flags.bad_for_scaling, all=False).count(
        False
    )
    assert len(set(refls["id"])) == 2
    assert len(set(refls["imageset_id"])) == 2
    for id_ in range(2):
        sel = refls["id"] == id_
        assert set(refls["imageset_id"].select(sel)) == {id_}

    # run again, optimising errors, and continuing from where last run left off.
    extra_args = [
        "error_model=basic",
        "unmerged_mtz=unmerged.mtz",
        "check_consistent_indexing=True",
    ]
    run_one_scaling(tmp_path, ["scaled.refl", "scaled.expt"] + extra_args)
    # Now inspect output, check it hasn't changed drastically, or if so verify
    # that the new behaviour is more correct and update test accordingly.
    # Note: error optimisation currently appears to give worse results here!
    result = get_merging_stats(tmp_path / "unmerged.mtz")
    expected_nobs = 5560  # 22/06/20
    print(result.overall.r_pim)
    print(result.overall.cc_one_half)
    assert abs(result.overall.n_obs - expected_nobs) < 100
    assert result.overall.r_pim < 0.016  # at #22/06/20, value was 0.015
    assert result.overall.cc_one_half > 0.997  # at #22/06/20, value was 0.999


def test_multi_scale_individual_error_models(dials_data, tmp_path):
    """Test standard scaling of two datasets."""

    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    refl_2 = data_dir / "25_integrated.pickle"
    expt_2 = data_dir / "25_integrated_experiments.json"
    extra_args = [
        "unmerged_mtz=unmerged.mtz",
        "error_model.grouping=individual",
        "intensity_choice=profile",
    ]

    run_one_scaling(tmp_path, [refl_1, refl_2, expt_1, expt_2] + extra_args)

    # Now inspect output, check it hasn't changed drastically, or if so verify
    # that the new behaviour is more correct and update test accordingly.
    result = get_merging_stats(tmp_path / "unmerged.mtz")
    expected_nobs = 5358  # 04/05/21
    assert abs(result.overall.n_obs - expected_nobs) < 30
    assert result.overall.r_pim < 0.015  # at 04/05/21, value was 0.013
    assert result.overall.cc_one_half > 0.9975  # at 04/05/21, value was 0.999

    # Test that different error models were determined for each sweep.
    scaling_models = load.experiment_list(
        tmp_path / "scaled.expt", check_format=False
    ).scaling_models()
    params_1 = scaling_models[0].configdict["error_model_parameters"]
    params_2 = scaling_models[1].configdict["error_model_parameters"]
    assert params_1[0] != params_2[0]
    assert params_1[1] != params_2[1]


def test_multi_scale_exclude_images(dials_data, tmp_path):
    """Test scaling of multiple dataset with image exclusion."""
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    refl_2 = data_dir / "25_integrated.pickle"
    expt_2 = data_dir / "25_integrated_experiments.json"
    # Expect this dataset to be given batches 1-1800 and 1901-3600
    # Try excluding last two hundred batches
    extra_args = [
        "error_model=None",
        "intensity_choice=profile",
        "outlier_rejection=simple",
        "exclude_images=0:1601:1800",
        "exclude_images=1:1501:1700",
    ]

    run_one_scaling(tmp_path, [refl_1, refl_2, expt_1, expt_2] + extra_args)

    refls = flex.reflection_table.from_file(tmp_path / "scaled.refl")
    d1 = refls.select(refls["id"] == 0)
    d2 = refls.select(refls["id"] == 1)
    nd1_scaled = d1.get_flags(d1.flags.scaled).count(True)
    # full sweep would have 2312, expect ~2060
    assert nd1_scaled < 2100
    assert nd1_scaled > 2000
    nd2_scaled = d2.get_flags(d2.flags.scaled).count(True)
    # full sweep would have 3210
    assert nd2_scaled < 2900
    assert nd2_scaled > 2800

    scaling_models = load.experiment_list(
        tmp_path / "scaled.expt", check_format=False
    ).scaling_models()
    assert scaling_models[0].configdict["valid_image_range"] == [1, 1600]
    assert scaling_models[1].configdict["valid_image_range"] == [1, 1500]
    assert [0, 160.0] == pytest.approx(scaling_models[0].configdict["valid_osc_range"])
    assert [-145.0, 5.0] == pytest.approx(
        scaling_models[1].configdict["valid_osc_range"]
    )

    # Run again, excluding some more from one run.
    extra_args = [
        "error_model=None",
        "intensity_choice=profile",
        "outlier_rejection=simple",
        "exclude_images=0:1401:1600",
    ]
    run_one_scaling(tmp_path, ["scaled.refl", "scaled.expt"] + extra_args)
    scaling_models = load.experiment_list(
        tmp_path / "scaled.expt", check_format=False
    ).scaling_models()
    assert scaling_models[0].configdict["valid_image_range"] == [1, 1400]
    assert scaling_models[1].configdict["valid_image_range"] == [1, 1500]
    assert [0, 140.0] == pytest.approx(scaling_models[0].configdict["valid_osc_range"])
    assert [-145.0, 5.0] == pytest.approx(
        scaling_models[1].configdict["valid_osc_range"]
    )

    refls = flex.reflection_table.from_file(tmp_path / "scaled.refl")
    d1 = refls.select(refls["id"] == 0)
    d2 = refls.select(refls["id"] == 1)
    nd1_scaled = d1.get_flags(d1.flags.scaled).count(True)
    # full sweep would have 2312, expect 1800
    assert nd1_scaled < 1850
    assert nd1_scaled > 1750
    nd2_scaled = d2.get_flags(d2.flags.scaled).count(True)
    # full sweep would have 3210, expect ~2850
    assert nd2_scaled < 2900
    assert nd2_scaled > 2800


def test_scale_handle_bad_dataset(dials_data, tmp_path):
    """Set command line parameters such that one dataset does not meet the
    criteria for inclusion in scaling. Check that this is excluded and the
    scaling job completes without failure."""
    location = dials_data("multi_crystal_proteinase_k", pathlib=True)
    command = [
        "dials.scale",
        "reflection_selection.method=intensity_ranges",
        "Isigma_range=90.0,1000",
    ]
    for i in range(1, 6):
        command.append(location / f"experiments_{i}.json")
        command.append(location / f"reflections_{i}.pickle")

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    reflections = flex.reflection_table.from_file(tmp_path / "scaled.refl")
    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    assert len(expts) == 4
    assert len(reflections.experiment_identifiers()) == 4


def test_targeted_scaling(dials_data, tmp_path):
    """Test the targeted scaling workflow."""
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    target_refl = location / "scaled_35.refl"
    target_expt = location / "scaled_35.expt"

    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    refl_2 = data_dir / "25_integrated.pickle"
    expt_2 = data_dir / "25_integrated_experiments.json"

    # Do targeted scaling, use this as a chance to test the KB model as well.
    extra_args = ["model=KB"]
    run_one_scaling(tmp_path, [target_refl, refl_1, target_expt, expt_1] + extra_args)

    scaled_exp = tmp_path / "scaled.expt"
    scaled_refl = tmp_path / "scaled.refl"
    experiments_list = load.experiment_list(scaled_exp, check_format=False)
    assert len(experiments_list.scaling_models()) == 2
    assert experiments_list.scaling_models()[0].id_ == "physical"
    assert experiments_list.scaling_models()[1].id_ == "KB"

    extra_args = ["model=KB", "only_target=True"]
    run_one_scaling(tmp_path, [refl_2, scaled_refl, expt_2, scaled_exp] + extra_args)
    experiments_list = load.experiment_list(
        tmp_path / "scaled.expt", check_format=False
    )
    assert len(experiments_list.scaling_models()) == 1
    assert experiments_list.scaling_models()[0].id_ == "KB"


def test_shared_absorption_surface(dials_data, tmp_path):
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"
    refl_2 = data_dir / "25_integrated.pickle"
    expt_2 = data_dir / "25_integrated_experiments.json"

    # Do targeted scaling, use this as a chance to test the KB model as well.
    extra_args = ["share.absorption=True"]
    run_one_scaling(tmp_path, [refl_1, expt_1, refl_2, expt_2] + extra_args)

    expts = load.experiment_list(tmp_path / "scaled.expt", check_format=False)
    assert (
        expts.scaling_models()[0].components["absorption"].parameters
        == expts.scaling_models()[1].components["absorption"].parameters
    )
    assert (
        expts.scaling_models()[0].components["absorption"].parameter_esds
        == expts.scaling_models()[1].components["absorption"].parameter_esds
    )


def test_incremental_scale_workflow(dials_data, tmp_path):
    """Try scale, cosym, scale, cosym, scale."""
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl_1 = data_dir / "20_integrated.pickle"
    expt_1 = data_dir / "20_integrated_experiments.json"

    run_one_scaling(tmp_path, [refl_1, expt_1])

    # test order also - first new file before scaled
    refl_2 = data_dir / "25_integrated.pickle"
    expt_2 = data_dir / "25_integrated_experiments.json"
    command = ["dials.cosym", refl_2, expt_2, "scaled.refl", "scaled.expt"]

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "symmetrized.expt").is_file()
    assert (tmp_path / "symmetrized.refl").is_file()

    # now try scaling again to check everything is okay
    run_one_scaling(tmp_path, ["symmetrized.refl", "symmetrized.expt"])

    # test order also - first scaled file then new file
    refl_2 = data_dir / "30_integrated.pickle"
    expt_2 = data_dir / "30_integrated_experiments.json"
    command = [
        "dials.cosym",
        "scaled.refl",
        "scaled.expt",
        refl_1,
        expt_2,
        "output.reflections=symmetrized.refl",
        "output.experiments=symmetrized.expt",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "symmetrized.expt").is_file()
    assert (tmp_path / "symmetrized.refl").is_file()

    # now try scaling again to check everything is okay
    run_one_scaling(tmp_path, ["symmetrized.refl", "symmetrized.expt"])


@pytest.mark.parametrize(
    ("mode", "parameter", "parameter_values"),
    [
        ("single", None, None),
        ("multi", "physical.absorption_correction", None),
        ("multi", "model", "physical array"),
    ],
)
def test_scale_cross_validate(dials_data, tmp_path, mode, parameter, parameter_values):
    """Test standard scaling of one dataset."""
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    refl = data_dir / "20_integrated.pickle"
    expt = data_dir / "20_integrated_experiments.json"
    extra_args = [
        f"cross_validation_mode={mode}",
        "nfolds=2",
        "full_matrix=0",
        "error_model=None",
    ]
    if parameter:
        extra_args += [f"parameter={parameter}"]
    if parameter_values:
        extra_args += [f"parameter_values={parameter_values}"]
    command = ["dials.scale", refl, expt] + extra_args
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr


def test_few_reflections(dials_data):
    """
    Test that dials.symmetry does something sensible if given few reflections.

    Use some example integrated data generated from two ten-image 1° sweeps.  These
    each contain a few dozen integrated reflections.

    Also test the behaviour of dials.merge and dials.report on the output.

    By suppressing the output from dials.scale and dials.report, we obviate the need to
    run in a temporary directory.

    Args:
        dials_data: DIALS custom Pytest fixture for access to test data.
    """
    # Get the input experiment lists and reflection tables.
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    experiments = ExperimentList.from_file(data_dir / "11_integrated.expt")
    experiments.extend(ExperimentList.from_file(data_dir / "23_integrated.expt"))
    refls = "11_integrated.refl", "23_integrated.refl"
    reflections = [flex.reflection_table.from_file(data_dir / refl) for refl in refls]

    # Get and use the default parameters for dials.scale, suppressing HTML output.
    scale_params = scale.phil_scope.fetch(
        source=phil.parse("output.html=None")
    ).extract()
    # Does what it says on the tin.  Run scaling.
    scaled_expt, scaled_refl = scale.run_scaling(scale_params, experiments, reflections)

    # Get and use the default parameters for dials.merge.
    merge_params = merge.phil_scope.fetch(source=phil.parse("")).extract()
    # Run dials.merge on the scaling output.
    merge.merge_data_to_mtz(merge_params, scaled_expt, [scaled_refl])

    # Get and use the default parameters for dials.report, suppressing HTML output.
    report_params = report.phil_scope.fetch(
        source=phil.parse("output.html=None")
    ).extract()
    # Get an Analyser object, which does the dials.report stuff.
    analyse = report.Analyser(
        report_params,
        grid_size=report_params.grid_size,
        centroid_diff_max=report_params.centroid_diff_max,
    )
    # Run dials.report on scaling output.
    analyse(scaled_refl, scaled_expt)



from __future__ import annotations

from dials.algorithms.background.gmodel.algorithm import BackgroundAlgorithm
from dials_algorithms_background_gmodel_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "BackgroundAlgorithm",
    "BackgroundModel",
    "Creator",
    "PolarTransform",
    "PolarTransformResult",
    "StaticBackgroundModel",
)



#!/usr/bin/env python
# -*- mode: python; coding: utf-8; indent-tabs-mode: nil; python-indent: 4 -*-

# Running bootstrap requires a minimum Python version of 2.7.

# To download this file:
# wget https://raw.githubusercontent.com/dials/dials/main/installer/bootstrap.py
# or
# curl https://raw.githubusercontent.com/dials/dials/main/installer/bootstrap.py > bootstrap.py

from __future__ import absolute_import, division, print_function

import argparse
import json
import multiprocessing.pool
import os
import platform
import re
import shutil
import socket as pysocket
import stat
import subprocess
import sys
import tarfile
import threading
import time
import warnings
import zipfile

try:  # Python 3
    from urllib.error import HTTPError, URLError
    from urllib.request import Request, urlopen
except ImportError:  # Python 2
    from urllib2 import HTTPError, Request, URLError, urlopen

# Clean environment for subprocesses
clean_env = {
    key: value
    for key, value in os.environ.items()
    if key not in ("PYTHONPATH", "LD_LIBRARY_PATH", "DYLD_LIBRARY_PATH")
}

devnull = open(os.devnull, "wb")  # to redirect unwanted subprocess output
allowed_ssh_connections = {}
concurrent_git_connection_limit = threading.Semaphore(5)

_prebuilt_cctbx_base = "2022.8"  # September 2022 release


def make_executable(filepath):
    if os.name == "posix":
        mode = os.stat(filepath).st_mode
        mode |= (mode & 0o444) >> 2  # copy R bits to X
        # r--r--r-- => 0o444
        os.chmod(filepath, mode)


def install_micromamba(python, include_cctbx):
    """Download and install Micromamba"""
    if sys.platform.startswith("linux"):
        conda_platform = "linux"
        conda_arch = "linux-64"
        member = "bin/micromamba"
    elif sys.platform == "darwin":
        conda_platform = "macos"
        member = "bin/micromamba"
        if platform.machine() == "arm64":
            conda_arch = "osx-arm64"
        else:
            conda_arch = "osx-64"
    elif os.name == "nt":
        conda_platform = "windows"
        member = "Library/bin/micromamba.exe"
        conda_arch = "win-64"
    else:
        raise NotImplementedError(
            "Unsupported platform %s / %s" % (os.name, sys.platform)
        )
    url = "https://micro.mamba.pm/api/micromamba/{0}/latest".format(conda_arch)
    mamba_prefix = os.path.realpath("micromamba")
    clean_env["MAMBA_ROOT_PREFIX"] = mamba_prefix
    mamba = os.path.join(mamba_prefix, member.split("/")[-1])
    print("Downloading {url}:".format(url=url), end=" ")
    result = download_to_file(url, os.path.join(mamba_prefix, "micromamba.tar.bz2"))
    if result in (0, -1):
        sys.exit("Micromamba download failed")
    with tarfile.open(
        os.path.join(mamba_prefix, "micromamba.tar.bz2"), "r:bz2"
    ) as tar, open(mamba, "wb") as fh:
        fh.write(tar.extractfile(member).read())
    make_executable(mamba)

    # verify micromamba works and check version
    conda_info = subprocess.check_output([mamba, "--version"], env=clean_env)
    if sys.version_info.major > 2:
        conda_info = conda_info.decode("latin-1")
    print("Using Micromamba version", conda_info.strip())

    # Identify packages required for environment
    env_dir = os.path.join("modules", "dials", ".conda-envs")
    # First, check to see if we have an architecture-and-python-specific environment file
    filename = os.path.join(env_dir, "{0}_py{1}.txt".format(conda_arch, python))
    if not os.path.isfile(filename):
        # Otherwise, check if we have an architecture-specific environment file
        filename = os.path.join(env_dir, "{0}.txt".format(conda_arch))
    if not os.path.isfile(filename):
        # Otherwise, use the platform-specific fallback
        filename = os.path.join(env_dir, conda_platform + ".txt")
    if not os.path.isfile(filename):
        raise RuntimeError(
            "The environment file {filename} is not available".format(filename=filename)
        )

    # install a new environment or update an existing one
    prefix = os.path.realpath("conda_base")
    if os.path.exists(prefix):
        command = "install"
        text_messages = ["Updating", "update of"]
    else:
        command = "create"
        text_messages = ["Installing", "installation into"]
    python_requirement = "conda-forge::python=%s.*" % python

    command_list = [
        mamba,
        "--no-env",
        "--no-rc",
        "--prefix",
        prefix,
        "--root-prefix",
        mamba_prefix,
        command,
        "--file",
        filename,
        "--yes",
        "--channel",
        "conda-forge",
        "--override-channels",
        python_requirement,
    ]
    if include_cctbx:
        command_list.append("cctbx-base=" + _prebuilt_cctbx_base)

    print(
        "{text} dials environment from {filename} with Python {python}".format(
            text=text_messages[0], filename=filename, python=python
        )
    )
    for retry in range(5):
        retry += 1
        try:
            run_command(
                command=command_list,
                workdir=".",
            )
        except Exception:
            print(
                """
*******************************************************************************
There was a failure in constructing the conda environment.
Attempt {retry} of 5 will start {retry} minute(s) from {t}.
*******************************************************************************
""".format(
                    retry=retry, t=time.asctime()
                )
            )
            time.sleep(retry * 60)
        else:
            break
    else:
        sys.exit(
            """
The conda environment could not be constructed. Please check that there is a
working network connection for downloading conda packages.
"""
        )
    print("Completed {text}:\n  {prefix}".format(text=text_messages[1], prefix=prefix))
    with open(os.path.join(prefix, ".condarc"), "w") as fh:
        fh.write(
            """
changeps1: False
channels:
  - conda-forge
""".lstrip()
        )


def install_miniconda(location):
    """Download and install Miniconda3"""
    if sys.platform.startswith("linux"):
        filename = "Miniconda3-latest-Linux-x86_64.sh"
    elif sys.platform == "darwin":
        filename = "Miniconda3-latest-MacOSX-x86_64.sh"
    elif os.name == "nt":
        filename = "Miniconda3-latest-Windows-x86_64.exe"
    else:
        raise NotImplementedError(
            "Unsupported platform %s / %s" % (os.name, sys.platform)
        )
    url = "https://repo.anaconda.com/miniconda/" + filename
    filename = os.path.join(location, filename)

    print("Downloading {url}:".format(url=url), end=" ")
    result = download_to_file(url, filename)
    if result in (0, -1):
        sys.exit("Miniconda download failed")

    # run the installer
    if os.name == "nt":
        command = [
            filename,
            "/InstallationType=JustMe",
            "/RegisterPython=0",
            "/AddToPath=0",
            "/S",
            "/D=" + location,
        ]
    else:
        command = ["/bin/sh", filename, "-b", "-u", "-p", location]

    print("Installing Miniconda")
    run_command(command=command, workdir=".")


def install_conda(python, include_cctbx):
    # Find relevant conda base installation
    conda_base = os.path.realpath("miniconda")
    if os.name == "nt":
        conda_exe = os.path.join(conda_base, "Scripts", "conda.exe")
    else:
        conda_exe = os.path.join(conda_base, "bin", "conda")

    # default environment file for users
    environment_file = os.path.join(
        os.path.expanduser("~"), ".conda", "environments.txt"
    )

    def get_environments():
        """Return a set of existing conda environment paths"""
        try:
            with open(environment_file) as f:
                paths = f.readlines()
        except IOError:
            paths = []
        environments = set(  # noqa; C401, Python 2.7 compatibility
            os.path.normpath(env.strip()) for env in paths if os.path.isdir(env.strip())
        )
        env_dirs = (
            os.path.join(conda_base, "envs"),
            os.path.join(os.path.expanduser("~"), ".conda", "envs"),
        )
        for env_dir in env_dirs:
            if os.path.isdir(env_dir):
                for d in os.listdir(env_dir):
                    d = os.path.join(env_dir, d)
                    if os.path.isdir(d):
                        environments.add(d)

        return environments

    if os.path.isdir(conda_base) and os.path.isfile(conda_exe):
        print("Using miniconda installation from", conda_base)
    else:
        print("Installing miniconda into", conda_base)
        install_miniconda(conda_base)

    # verify consistency and check conda version
    if not os.path.isfile(conda_exe):
        sys.exit("Conda executable not found at " + conda_exe)

    environments = get_environments()

    conda_info = subprocess.check_output([conda_exe, "info", "--json"], env=clean_env)
    if sys.version_info.major > 2:
        conda_info = conda_info.decode("latin-1")
    conda_info = json.loads(conda_info)
    for env in environments:
        if env not in conda_info["envs"]:
            print("Consistency check:", env, "not in environments:")
            print(conda_info["envs"])
            warnings.warn(
                """
There is a mismatch between the conda settings in your home directory
and what "conda info" is reporting. This is not a fatal error, but if
an error is encountered, please check that your conda installation and
environments exist and are working.
""",
                RuntimeWarning,
            )

    # identify packages required for environment
    if os.name == "nt":
        conda_platform = "windows"
    elif sys.platform == "darwin":
        conda_platform = "macos"
    else:
        conda_platform = "linux"
    filename = os.path.join(
        "modules",
        "dials",
        ".conda-envs",
        "{platform}.txt".format(platform=conda_platform),
    )
    if not os.path.isfile(filename):
        raise RuntimeError(
            "The file {filename} is not available".format(filename=filename)
        )

    python_requirement = "conda-forge::python=%s.*" % python

    # make a new environment directory
    prefix = os.path.realpath("conda_base")

    # install a new environment or update and existing one
    if prefix in environments:
        command = "install"
        text_messages = ["Updating", "update of"]
    else:
        command = "create"
        text_messages = ["Installing", "installation into"]
    command_list = [
        conda_exe,
        command,
        "--prefix",
        prefix,
        "--file",
        filename,
        "--yes",
        "--channel",
        "conda-forge",
        "--override-channels",
        python_requirement,
    ]
    if include_cctbx:
        command_list.append("cctbx-base=" + _prebuilt_cctbx_base)
    if os.name == "nt":
        command_list = [
            "cmd.exe",
            "/C",
            " ".join(
                [os.path.join(conda_base, "Scripts", "activate"), "base", "&&"]
                + command_list
            )
            .replace("<", "^<")
            .replace(">", "^>"),
        ]
    print(
        "{text} dials environment from {filename} with Python {python}".format(
            text=text_messages[0], filename=filename, python=python
        )
    )
    for retry in range(5):
        retry += 1
        try:
            run_command(
                command=command_list,
                workdir=".",
            )
        except Exception:
            print(
                """
*******************************************************************************
There was a failure in constructing the conda environment.
Attempt {retry} of 5 will start {retry} minute(s) from {t}.
*******************************************************************************
""".format(
                    retry=retry, t=time.asctime()
                )
            )
            time.sleep(retry * 60)
        else:
            break
    else:
        sys.exit(
            """
The conda environment could not be constructed. Please check that there is a
working network connection for downloading conda packages.
"""
        )
    print("Completed {text}:\n  {prefix}".format(text=text_messages[1], prefix=prefix))
    with open(os.path.join(prefix, ".condarc"), "w") as fh:
        fh.write(
            """
changeps1: False
channels:
  - conda-forge
""".lstrip()
        )


def run_command(command, workdir):
    print("Running %s (in %s)" % (" ".join(command), workdir))
    workdir = os.path.abspath(workdir)
    try:
        os.makedirs(workdir)
    except OSError:
        pass
    try:
        p = subprocess.Popen(args=command, cwd=workdir, env=clean_env)
    except Exception as e:
        if isinstance(e, OSError):
            if e.errno == 2:
                executable = os.path.normpath(os.path.join(workdir, command[0]))
                raise RuntimeError("Could not run %s: File not found" % executable)
        if "child_traceback" in dir(e):
            print("Calling subprocess resulted in error; ", e.child_traceback)
        raise e

    try:
        p.wait()
    except KeyboardInterrupt:
        print("\nReceived CTRL+C, trying to stop subprocess...\n")
        p.terminate()
        raise
    if p.returncode:
        sys.exit("Process failed with return code %s" % p.returncode)


def run_indirect_command(command, args):
    print("(via conda environment) " + command)
    try:
        os.mkdir("build")
    except OSError:
        pass
    if os.name == "nt":
        filename = os.path.join("build", "indirection.cmd")
        with open(filename, "w") as fh:
            fh.write("call %s\\conda_base\\condabin\\activate.bat\r\n" % os.getcwd())
            fh.write("shift\r\n")
            fh.write("%*\r\n")
        if not command.endswith((".bat", ".cmd", ".exe")):
            command = command + ".bat"
        indirection = ["cmd.exe", "/C", "indirection.cmd"]
    else:
        filename = os.path.join("build", "indirection.sh")
        with open(filename, "w") as fh:
            fh.write("#!/bin/bash\n")
            fh.write("source %s/conda_base/etc/profile.d/conda.sh\n" % os.getcwd())
            fh.write("conda activate %s/conda_base\n" % os.getcwd())
            fh.write('"$@"\n')
        make_executable(filename)
        indirection = ["./indirection.sh"]
    run_command(
        command=indirection + [command] + args,
        workdir="build",
    )


def download_to_file(url, file, quiet=False):
    """Downloads a URL to file. Returns the file size.
    Returns -1 if the downloaded file size does not match the expected file
    size
    Returns -2 if the download is skipped due to the file at the URL not
    being newer than the local copy (identified by matching timestamp and
    size)
    """

    # Create directory structure if necessary
    if os.path.dirname(file):
        try:
            os.makedirs(os.path.dirname(file))
        except Exception:
            pass

    localcopy = os.path.isfile(file)

    try:
        from ssl import SSLError
    except ImportError:
        SSLError = None

    # Open connection to remote server
    try:
        url_request = Request(url)
        if localcopy:
            # Shorten timeout to 7 seconds if a copy of the file is already present
            socket = urlopen(url_request, None, 7)
        else:
            socket = urlopen(url_request)
    except SSLError as e:
        # This could be a timeout
        if localcopy:
            # Download failed for some reason, but a valid local copy of
            # the file exists, so use that one instead.
            if not quiet:
                print(str(e))
            return -2
        # otherwise pass on the error message
        raise
    except (pysocket.timeout, HTTPError) as e:
        if localcopy:
            # Download failed for some reason, but a valid local copy of
            # the file exists, so use that one instead.
            if not quiet:
                print(str(e))
            return -2
        # otherwise pass on the error message
        raise
    except URLError as e:
        if localcopy:
            # Download failed for some reason, but a valid local copy of
            # the file exists, so use that one instead.
            if not quiet:
                print(str(e))
            return -2
        # if url fails to open, try using curl
        # temporary fix for old OpenSSL in system Python on macOS
        # https://github.com/cctbx/cctbx_project/issues/33
        command = ["/usr/bin/curl", "-fLo", file, "--retry", "5", url]
        subprocess.call(command)
        socket = None  # prevent later socket code from being run
        try:
            received = os.path.getsize(file)
        except OSError:
            raise RuntimeError("Download failed")

    if socket is not None:
        try:
            file_size = int(socket.info().get("Content-Length"))
        except Exception:
            file_size = 0

        remote_mtime = 0
        try:
            remote_mtime = time.mktime(socket.info().getdate("last-modified"))
        except Exception:
            pass

        if file_size > 0:
            if remote_mtime > 0:
                # check if existing file matches remote size and timestamp
                try:
                    (
                        mode,
                        ino,
                        dev,
                        nlink,
                        uid,
                        gid,
                        size,
                        atime,
                        mtime,
                        ctime,
                    ) = os.stat(file)
                    if (size == file_size) and (remote_mtime == mtime):
                        if not quiet:
                            print("local copy is current")
                        socket.close()
                        return -2
                except Exception:
                    # proceed with download if timestamp/size check fails for any reason
                    pass

            hr_size = (file_size, "B")
            if hr_size[0] > 500:
                hr_size = (hr_size[0] / 1024, "kB")
            if hr_size[0] > 500:
                hr_size = (hr_size[0] / 1024, "MB")
            if not quiet:
                print("%.1f %s" % hr_size)
                print("    [0%", end="")
                sys.stdout.flush()  # becomes print(flush=True) when we move to 3.3+

        received = 0
        block_size = 8192
        progress = 1
        # Write to the file immediately so we can empty the buffer
        tmpfile = file + ".tmp"

        with open(tmpfile, "wb") as fh:
            while True:
                block = socket.read(block_size)
                received += len(block)
                fh.write(block)
                if file_size > 0 and not quiet:
                    while (100 * received / file_size) > progress:
                        progress += 1
                        if (progress % 20) == 0:
                            print(progress, end="%")
                            sys.stdout.flush()  # becomes print(flush=True) when we move to 3.3+
                        elif (progress % 2) == 0:
                            print(".", end="")
                            sys.stdout.flush()  # becomes print(flush=True) when we move to 3.3+
                if not block:
                    break
        socket.close()

        if not quiet:
            if file_size > 0:
                print("]")
            else:
                print("%d kB" % (received / 1024))
            sys.stdout.flush()  # becomes print(flush=True) when we move to 3.3+

        # Do not overwrite file during the download. If a download temporarily fails we
        # may still have a clean, working (yet older) copy of the file.
        shutil.move(tmpfile, file)

        if (file_size > 0) and (file_size != received):
            return -1

        if remote_mtime > 0:
            # set file timestamp if timestamp information is available
            st = os.stat(file)
            atime = st[stat.ST_ATIME]  # current access time
            os.utime(file, (atime, remote_mtime))

    return received


def unzip(archive, directory, trim_directory=0):
    """unzip a file into a directory."""
    if not zipfile.is_zipfile(archive):
        raise Exception(
            "Cannot install %s: %s is not a valid .zip file" % (directory, archive)
        )
    z = zipfile.ZipFile(archive, "r")
    for member in z.infolist():
        is_directory = member.filename.endswith("/")
        filename = os.path.join(*member.filename.split("/")[trim_directory:])
        if filename != "":
            filename = os.path.normpath(filename)
            if "../" in filename:
                raise Exception(
                    "Archive %s contains invalid filename %s" % (archive, filename)
                )
            filename = os.path.join(directory, filename)
            upperdirs = os.path.dirname(filename)
            try:
                if is_directory and not os.path.exists(filename):
                    os.makedirs(filename)
                elif upperdirs and not os.path.exists(upperdirs):
                    os.makedirs(upperdirs)
            except Exception:
                pass
            if not is_directory:
                with z.open(member) as source:
                    with open(filename, "wb") as target:
                        shutil.copyfileobj(source, target)

                # Preserve executable permission, if set
                unix_executable = member.external_attr >> 16 & 0o111
                # rwxrwxrwx => --x--x--x => 0o111
                if unix_executable:
                    make_executable(filename)
    z.close()


def set_git_repository_config_to_rebase(config):
    with open(config, "r") as fh:
        cfg = fh.readlines()

    branch, remote, rebase = False, False, False
    insertions = []
    for n, line in enumerate(cfg):
        if line.startswith("["):
            if branch and remote and not rebase:
                insertions.insert(0, (n, branch))
            if line.startswith("[branch"):
                branch = line.split('"')[1]
            else:
                branch = False
            remote, rebase = False, False
        if re.match(r"remote\s*=", line.strip()):
            remote = True
        if re.match(r"rebase\s*=", line.strip()):
            rebase = True
    if branch and remote and not rebase:
        insertions.insert(0, (n + 1, branch))
    for n, branch in insertions:
        cfg.insert(n, "\trebase = true\n")
    with open(config, "w") as fh:
        fh.write("".join(cfg))


def git(module, git_available, ssh_available, reference_base, settings):
    """Retrieve a git repository, either by running git directly
    or by downloading and unpacking an archive.
    """
    destination = os.path.join("modules", module)

    if os.path.exists(destination):
        if os.path.isfile(os.path.join(destination, ".git")):
            return module, "WARNING", "Existing git worktree directory -- skipping"
        if not os.path.isdir(os.path.join(destination, ".git")):
            return module, "WARNING", "Existing non-git directory -- skipping"
        if not git_available:
            return module, "WARNING", "Cannot update module, git command not found"

        with open(os.path.join(destination, ".git", "HEAD"), "r") as fh:
            if fh.read(4) != "ref:":
                return (
                    module,
                    "WARNING",
                    "Cannot update existing git repository! You are not on a branch.\n"
                    "This may be legitimate when run eg. via Jenkins, but be aware that you cannot commit any changes",
                )

        with concurrent_git_connection_limit:
            p = subprocess.Popen(
                args=["git", "pull", "--rebase"],
                cwd=destination,
                env=clean_env,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            # This may fail for unclean trees and merge problems. In this case manual
            # user intervention will be required.
            # For the record, you can clean up the tree and *discard ALL changes* with
            #   git reset --hard origin/master
            #   git clean -dffx
            try:
                output, _ = p.communicate()
                output = output.decode("latin-1")
            except KeyboardInterrupt:
                print("\nReceived CTRL+C, trying to terminate subprocess...\n")
                p.terminate()
                raise
        if p.returncode:
            return (
                module,
                "WARNING",
                "Cannot update existing git repository! Unclean tree or merge problems.\n"
                + output,
            )
        # Show the hash for the checked out commit for debugging purposes
        p = subprocess.Popen(
            args=["git", "rev-parse", "HEAD", "--abbrev-ref", "HEAD"],
            cwd=destination,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        output, _ = p.communicate()
        output = output.decode("latin-1")
        if p.returncode:
            return module, "WARNING", "Cannot get git repository revision\n" + output
        output = output.split()
        if len(output) == 2:
            return module, "OK", "Checked out revision %s (%s)" % (output[0], output[1])
        return module, "OK", "Checked out revision " + output[0].strip()

    try:
        os.makedirs("modules")
    except OSError:
        pass

    remote_branch = settings.get("branch-remote", settings["branch-local"])

    if not git_available:
        # Fall back to downloading a static archive
        url = "https://github.com/%s/archive/%s.zip" % (
            settings.get("effective-repository", settings.get("base-repository")),
            remote_branch,
        )
        filename = os.path.join("modules", "%s-%s.zip" % (module, remote_branch))
        try:
            download_to_file(url, filename, quiet=True)
        except Exception:
            print("Error downloading", url)
            raise
        unzip(filename, destination, trim_directory=1)
        return module, "OK", "Downloaded branch %s from static archive" % remote_branch

    if ssh_available:
        remote_pattern = "git@github.com:%s.git"
    else:
        remote_pattern = "https://github.com/%s.git"

    if git_available:
        # Determine the git version
        p = subprocess.Popen(
            ["git", "--version"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        output, _ = p.communicate()
        output = output.decode("latin-1")
        parts = output.split(" ", 2)
        if p.returncode or not parts[:2] == ["git", "version"]:
            raise RuntimeError("Could not determine git version")
        # Version comes in:
        #    "git version x.y.z"
        # or "git version x.y.z.windows.n"
        git_version = tuple(int(x) if x.isnumeric() else x for x in parts[2].split("."))

    secondary_remote = settings.get("effective-repository") and (
        settings["effective-repository"] != settings.get("base-repository")
    )
    direct_branch_checkout = []
    if not secondary_remote and remote_branch == settings["branch-local"]:
        direct_branch_checkout = ["-b", remote_branch]
    reference_parameters = []
    if reference_base and os.path.exists(os.path.join(reference_base, module, ".git")):
        # Use -if-able so that we don't have errors over unreferenced submodules
        reference_type = "--reference-if-able"
        if git_version < (2, 11, 0):
            # As a fallback, use the old parameter. This will fail if
            # there are submodules and the reference does not have all
            # the required submodules
            reference_type = "--reference"
        reference_parameters = [
            reference_type,
            os.path.join(reference_base, module),
        ]

    with concurrent_git_connection_limit:
        p = subprocess.Popen(
            args=["git", "clone", "--recursive"]
            + direct_branch_checkout
            + [
                remote_pattern
                % settings.get("base-repository", settings.get("effective-repository")),
                module,
            ]
            + reference_parameters,
            cwd="modules",
            env=clean_env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        try:
            output, _ = p.communicate()
            output = output.decode("latin-1")
        except KeyboardInterrupt:
            print("\nReceived CTRL+C, trying to terminate subprocess...\n")
            p.terminate()
            raise
        if p.returncode:
            return (module, "ERROR", "Cannot checkout git repository\n" + output)

    if reference_parameters:
        # Sever the link between checked out and reference repository
        returncode = subprocess.call(
            ["git", "repack", "-a", "-d"],
            cwd=destination,
            env=clean_env,
            stdout=devnull,
            stderr=devnull,
        )
        if returncode:
            return (
                module,
                "ERROR",
                "Could not detach git repository from reference. Repository may be in invalid state!\n"
                "Run 'git repack -a -d' in the repository, or delete and recreate it",
            )
        alternates = os.path.join(destination, ".git", "objects", "info", "alternates")
        if os.path.exists(alternates):
            os.remove(alternates)

    if secondary_remote:
        returncode = subprocess.call(
            [
                "git",
                "remote",
                "add",
                "upstream",
                remote_pattern % settings["effective-repository"],
            ],
            cwd=destination,
            env=clean_env,
            stdout=devnull,
            stderr=devnull,
        )
        if returncode:
            return (
                module,
                "ERROR",
                "Could not add upstream remote to repository. Repository may be in invalid state!",
            )
        with concurrent_git_connection_limit:
            returncode = subprocess.call(
                ["git", "fetch", "upstream"],
                cwd=destination,
                env=clean_env,
                stdout=devnull,
                stderr=devnull,
            )
        if returncode:
            return (
                module,
                "ERROR",
                "Could not fetch upstream repository %s. Repository may be in invalid state!"
                % settings["effective-repository"],
            )

    set_git_repository_config_to_rebase(os.path.join(destination, ".git", "config"))

    if not direct_branch_checkout:
        # set up the local branch with tracking
        returncode = subprocess.call(
            [
                "git",
                "checkout",
                "-B",
                settings["branch-local"],
                "--track",
                "%s/%s" % ("upstream" if secondary_remote else "origin", remote_branch),
            ],
            cwd=destination,
            env=clean_env,
            stdout=devnull,
            stderr=devnull,
        )
        if returncode:
            return (
                module,
                "ERROR",
                "Could not check out alternate branch %s. Repository may be in invalid state!"
                % remote_branch,
            )

        if secondary_remote:
            # When checking out a branch from a secondary repository under a
            # different name set up a git pre-commit hook to write protect
            # this branch.
            hookfile = os.path.join(destination, ".git", "hooks", "pre-commit")
            with open(hookfile, "w") as fh:
                fh.write(
                    """
#!/bin/sh

#
# Reject commits to '{branch}' to avoid accidental
# commits to the alternate repository {repository}
#

if [ "$(git rev-parse --abbrev-ref HEAD)" == "{branch}" ]
then
    echo "Please do not commit to the '{branch}' branch."
    echo "You can create a new branch or commit directly to master."
    echo
    exit 1
fi
""".format(
                        branch=settings["branch-local"],
                        repository=settings["effective-repository"],
                    )
                )
            make_executable(hookfile)

    # Show the hash for the checked out commit for debugging purposes
    p = subprocess.Popen(
        args=["git", "rev-parse", "HEAD"],
        cwd=destination,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    output, _ = p.communicate()
    output = output.decode("latin-1")
    if p.returncode:
        return (
            module,
            "WARNING",
            "Cannot get git repository revision\n" + output,
        )
    git_status = settings["branch-local"]
    if settings["branch-local"] != remote_branch:
        git_status += " tracking " + remote_branch
    if secondary_remote:
        git_status += " at " + settings["effective-repository"]
    return module, "OK", "Checked out revision %s (%s)" % (output.strip(), git_status)


def update_sources(options):
    if options.git_reference:
        reference_base = os.path.abspath(os.path.expanduser(options.git_reference))
    else:
        if os.name == "posix" and pysocket.gethostname().endswith(".diamond.ac.uk"):
            reference_base = (
                "/dls/science/groups/scisoft/DIALS/repositories/git-reference"
            )
        else:
            reference_base = None
    try:
        git_available = not subprocess.call(
            ["git", "--version"], stdout=devnull, stderr=devnull
        )
    except OSError:
        git_available = False
    ssh_available = False
    if git_available:
        try:
            returncode = subprocess.call(
                [
                    "ssh",
                    "-oBatchMode=yes",
                    "-oStrictHostKeyChecking=no",
                    "-T",
                    "git@github.com",
                ],
                stdout=devnull,
                stderr=devnull,
            )
            # SSH errors lead to 255
            ssh_available = returncode in (0, 1)
        except OSError:
            pass

    repositories = {
        source.split("/")[1]: {"base-repository": source, "branch-local": "dials-3.12"}
        for source, branch in (
            ("dials/annlib_adaptbx", "master"),
            ("dials/dxtbx", "main"),
            ("dials/annlib", "master"),
            ("dials/cbflib", "main"),
            ("dials/ccp4io", "master"),
            ("dials/ccp4io_adaptbx", "master"),
            ("dials/cctbx_project", "master"),
            ("dials/dials", "main"),
            ("dials/gui_resources", "master"),
            ("xia2/xia2", "main"),
        )
    }
    if options.prebuilt_cctbx:
        repositories["cctbx_project"]["branch-local"] = (
            "releases/" + _prebuilt_cctbx_base
        )
    else:
        repositories["cctbx_project"] = {
            "base-repository": "dials/cctbx_project",
            "branch-local": "dials-3.12",
        }

    for source, setting in options.branch:
        if source not in repositories:
            sys.exit("Unknown repository %s" % source)
        setting = re.match(
            r"^(?:(\w+/\w+)@?)?([a-zA-Z0-9._\-]+)?(?::([a-zA-Z0-9._\-]+))?$", setting
        )
        if not setting:
            sys.exit("Could not parse the branch setting for repository %s" % source)
        _repository, _branch_remote, _branch_local = setting.groups()
        if _repository:
            repositories[source] = {
                "base-repository": _repository,
                "branch-remote": _branch_remote or "master",
                "branch-local": _branch_local or _branch_remote or "master",
            }
        elif _branch_remote:
            repositories[source]["branch-remote"] = _branch_remote
            repositories[source]["branch-local"] = _branch_local or _branch_remote
        elif _branch_local:
            repositories[source]["branch-local"] = _branch_local

    def _git_fn(repository):
        return git(
            repository,
            git_available,
            ssh_available,
            reference_base,
            repositories[repository],
        )

    success = True
    update_pool = multiprocessing.pool.ThreadPool(20)
    try:
        for result in update_pool.imap_unordered(_git_fn, repositories):
            module, result, output = result
            output = (result + " - " + output).replace(
                "\n", "\n" + " " * (len(module + result) + 5)
            )
            if result == "ERROR":
                success = False
            if os.name == "posix" and sys.stdout.isatty():
                if result == "OK":
                    output = "\x1b[32m" + output + "\x1b[0m"
                elif result == "WARNING":
                    output = "\x1b[33m" + output + "\x1b[0m"
                elif result == "ERROR":
                    output = "\x1b[31m" + output + "\x1b[0m"
            print(module + ": " + output)
    except KeyboardInterrupt:
        update_pool.terminate()
        sys.exit("\naborted with Ctrl+C")
    except Exception:
        update_pool.terminate()
        raise
    update_pool.close()
    update_pool.join()
    if not success:
        sys.exit("\nFailed to update one or more repositories")


def run_tests():
    dispatch_extension = ".bat" if os.name == "nt" else ""
    print("Running dxtbx tests")
    run_command(
        [
            os.path.join(
                "..", "..", "build", "bin", "libtbx.pytest" + dispatch_extension
            ),
            "--regression",
            "-n",
            "auto",
        ],
        workdir=os.path.join("modules", "dxtbx"),
    )
    print("Running dials tests")
    run_command(
        [
            os.path.join(
                "..", "..", "build", "bin", "libtbx.pytest" + dispatch_extension
            ),
            "--regression",
            "-n",
            "auto",
        ],
        workdir=os.path.join("modules", "dials"),
    )


def refresh_build(prebuilt_cctbx):
    print("Running libtbx.refresh")
    run_indirect_command(
        "libtbx.refresh" if prebuilt_cctbx else os.path.join("bin", "libtbx.refresh"),
        args=[],
    )


def install_precommit():
    print("Installing precommits")
    run_indirect_command(
        os.path.join("bin", "libtbx.precommit"),
        args=["install"],
    )


def configure_build(config_flags, prebuilt_cctbx):
    if os.name == "nt":
        conda_python = os.path.join(os.getcwd(), "conda_base", "python.exe")
    elif sys.platform.startswith("darwin"):
        conda_python = os.path.join(
            "..", "conda_base", "python.app", "Contents", "MacOS", "python"
        )
    else:
        conda_python = os.path.join("..", "conda_base", "bin", "python")

    # write a new-style environment setup script
    with open(("dials.bat" if os.name == "nt" else "dials"), "w"):
        pass  # ensure we write a new-style environment setup script

    if os.name != "nt" and not any(
        flag.startswith("--compiler=") for flag in config_flags
    ):
        config_flags.append("--compiler=conda")
    if "--enable_cxx11" not in config_flags:
        config_flags.append("--enable_cxx11")
    if "--use_conda" not in config_flags:
        config_flags.append("--use_conda")

    print("Setting up build directory")
    if prebuilt_cctbx:
        run_indirect_command(
            command="libtbx.configure",
            args=["cbflib", "dials", "dxtbx", "prime", "xia2"],
        )
        return

    configcmd = [
        os.path.join("..", "modules", "cctbx_project", "libtbx", "configure.py"),
        "cctbx",
        "scitbx",
        "libtbx",
        "iotbx",
        "mmtbx",
        "smtbx",
        "gltbx",
        "wxtbx",
        "cbflib",
        "dxtbx",
        "xfel",
        "dials",
        "xia2",
        "prime",
        "--skip_phenix_dispatchers",
    ] + config_flags

    run_indirect_command(
        command=conda_python,
        args=configcmd,
    )


def make_build(prebuilt_cctbx):
    try:
        nproc = len(os.sched_getaffinity(0))
    except AttributeError:
        nproc = multiprocessing.cpu_count()
    if prebuilt_cctbx:
        command = "libtbx.scons"
    else:
        command = os.path.join("bin", "libtbx.scons")

    run_indirect_command(command, args=["-j", str(nproc)])
    if not prebuilt_cctbx:
        # run build again to make sure everything is built
        run_indirect_command(command, args=["-j", str(nproc)])


def repository_at_tag(string):
    try:
        repository, tag = string.split("@", 1)
        return (repository, tag)
    except ValueError:
        raise argparse.ArgumentTypeError(
            "%s does not follow the repository@branch format" % string
        )


class Choices(tuple):
    # Python bug https://bugs.python.org/issue27227, https://bugs.python.org/issue9625
    def __new__(cls, *args, **kwargs):
        x = tuple.__new__(cls, *args, **kwargs)
        Choices.__init__(x, *args, **kwargs)
        return x

    def __init__(self, *args, **kwargs):
        self.default = []

    def __contains__(self, item):
        return tuple.__contains__(self, item) or item is self.default


def run():
    prog = os.environ.get("LIBTBX_DISPATCHER_NAME")
    if prog is None or prog.startswith("python") or prog.endswith("python"):
        prog = os.path.basename(sys.argv[0])

    description = """
  You may specify one or more actions:
    update - Update source repositories (cctbx, cbflib, etc.)
    base - Create conda environment
    build - Build
    test - Run tests

  The default actions are: update, base, build

  Example:

    python bootstrap.py update base build test
  """

    parser = argparse.ArgumentParser(
        prog=prog,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    action_choices = Choices(("update", "base", "build", "test"))
    action_choices.default = ["update", "base", "build"]
    parser.add_argument(
        "actions",
        nargs="*",
        help="actions for building as described above",
        choices=action_choices,
        default=action_choices.default,
    )
    parser.add_argument(
        "--git-reference",
        help="Path to a directory containing reference copies of repositories for faster checkouts.",
    )
    parser.add_argument(
        "--config-flags",
        help="""Pass flags to the configuration step. Flags should
be passed separately with quotes to avoid confusion (e.g
--config_flags="--build=debug" --config_flags="--another_flag")""",
        action="append",
        default=[],
    )
    parser.add_argument(
        "--python",
        help="Install this minor version of Python (default: %(default)s)",
        default="3.9",
        choices=("3.8", "3.9", "3.10"),
    )
    parser.add_argument(
        "--branch",
        type=repository_at_tag,
        action="append",
        default=[],
        help=(
            "during 'update' step when a repository is newly cloned set it to a given branch."
            "Specify as repository@branch, eg. 'dials@dials-next'"
        ),
    )
    parser.add_argument(
        "--conda",
        help="Use miniconda instead of micromamba for the base installation step",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--clean",
        help="Remove temporary conda environments and package caches after installation",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        # Use the conda-forge cctbx package instead of compiling cctbx from scratch
        # This is not currently supported outside of CI builds
        "--prebuilt-cctbx",
        help=argparse.SUPPRESS,
        default=False,
        action="store_true",
    )

    options = parser.parse_args()

    print("Performing actions:", " ".join(options.actions))

    # Add sources
    if "update" in options.actions:
        update_sources(options)

    # Build base packages
    if "base" in options.actions:
        if options.conda:
            install_conda(options.python, include_cctbx=options.prebuilt_cctbx)
            if options.clean:
                shutil.rmtree(os.path.realpath("miniconda"))
        else:
            install_micromamba(options.python, include_cctbx=options.prebuilt_cctbx)
            if options.clean:
                shutil.rmtree(os.path.realpath("micromamba"))

    # Configure, make, get revision numbers
    if "build" in options.actions:
        configure_build(options.config_flags, prebuilt_cctbx=options.prebuilt_cctbx)
        make_build(prebuilt_cctbx=options.prebuilt_cctbx)
        refresh_build(prebuilt_cctbx=options.prebuilt_cctbx)
        install_precommit()

    # Tests, tests
    if "test" in options.actions:
        run_tests()

    print("\nBootstrap success: %s" % ", ".join(options.actions))


if __name__ == "__main__":
    run()



from __future__ import annotations

import json
import os.path
import pathlib

import procrunner
import pytest

from dxtbx.serialize import load

from dials.command_line.ssx_index import run


@pytest.mark.xfel
def test_ssx_index_reference_geometry(dials_data, tmp_path):
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    expts = ssx / "imported_with_ref_5.expt"
    refls = ssx / "strong_5.refl"
    pathlib.Path.mkdir(tmp_path / "nuggets")
    result = procrunner.run(
        ["dials.ssx_index", expts, refls, "output.nuggets=nuggets", "min_spots=72"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "indexed.refl").is_file()
    assert (tmp_path / "indexed.expt").is_file()
    assert (tmp_path / "dials.ssx_index.html").is_file()
    experiments = load.experiment_list(tmp_path / "indexed.expt", check_format=False)
    assert len(experiments) == 3  # only 3 out of the 5 get indexed
    for i in [0, 1, 2, 4]:
        assert tmp_path.joinpath(
            f"nuggets/nugget_index_merlin0047_1700{i}.cbf.json"
        ).is_file()
    filtered_json = tmp_path.joinpath("nuggets/nugget_index_filtered_images.json")
    assert filtered_json.is_file()
    with open(filtered_json) as f:
        data = json.load(f)
    assert data["filtered_images"] == [4]


@pytest.mark.xfel
def test_ssx_index_no_reference_geometry(dials_data, tmp_path):
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    expts = ssx / "imported_no_ref_5.expt"
    refls = ssx / "strong_5.refl"

    result = procrunner.run(
        ["dials.ssx_index", expts, refls, "-vv"],
        working_directory=tmp_path,
    )

    assert not result.returncode and not result.stderr
    assert (tmp_path / "indexed.refl").is_file()
    assert (tmp_path / "indexed.expt").is_file()
    assert (tmp_path / "dials.ssx_index.html").is_file()
    experiments = load.experiment_list(tmp_path / "indexed.expt", check_format=False)
    assert (
        len(experiments) == 3
    )  # only 3 out of the 5 get indexed if no reference geometry


def test_ssx_index_bad_input(dials_data, run_in_tmp_path):
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    expts = str(ssx / "imported_no_ref_5.expt")
    refls = str(ssx / "strong_1.refl")

    run([expts, refls])
    assert os.path.exists("indexed.refl")
    assert os.path.exists("indexed.expt")
    experiments = load.experiment_list("indexed.expt", check_format=False)
    assert len(experiments) == 0


@pytest.mark.xfel
def test_ssx_index_input_unit_cell(dials_data, run_in_tmp_path):
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    expts = str(ssx / "imported_with_ref_5.expt")
    refls = str(ssx / "strong_5.refl")

    # invoke the run function
    run([expts, refls, "max_lattices=2", "unit_cell=96.4,96.4,96.4,90,90,90"])

    assert os.path.exists("indexed.refl")
    assert os.path.exists("indexed.expt")
    assert os.path.exists("dials.ssx_index.html")
    experiments = load.experiment_list("indexed.expt", check_format=False)
    assert (
        len(experiments) == 5
    )  # only 4 out of the 5 images gets indexed, one has two lattices

    # now if just fft1d method:
    # invoke the run function
    run([expts, refls, "method=fft1d", "unit_cell=96.4,96.4,96.4,90,90,90"])
    experiments = load.experiment_list("indexed.expt", check_format=False)
    assert (
        len(experiments) == 2
    )  # only 2 out of the 5 images get indexed without real space grid search



"""Shared functions for tests."""
from __future__ import annotations

from unittest.mock import Mock

from dials.array_family import flex


def mock_reflection_file_object(id_=0, identifier=True):
    """Create a mock reflection_file_object."""
    fileobj = Mock()
    r = flex.reflection_table()
    r["id"] = flex.int([-1, id_, id_])
    if identifier:
        r.experiment_identifiers()[id_] = str(id_)
    fileobj.data = r
    return fileobj


def mock_two_reflection_file_object(ids=(0, 2)):
    """Create a mock reflection_file_object with two datasets."""
    fileobj = Mock()
    r = flex.reflection_table()
    r["id"] = flex.int([-1, ids[0], ids[0], ids[1], ids[1]])
    r.experiment_identifiers()[ids[0]] = str(ids[0])
    r.experiment_identifiers()[ids[1]] = str(ids[1])
    fileobj.data = r
    return fileobj



from __future__ import annotations

import pytest

from dials.algorithms.indexing.basis_vector_search import (
    FFT1D,
    FFT3D,
    RealSpaceGridSearch,
)


class TestStrategies:
    def check_results(self, unit_cell, basis_vectors):
        # check we have found a basis vector corresponding to each unit cell parameter
        for p in unit_cell.parameters()[:3]:
            found = False
            for v in basis_vectors:
                if p == pytest.approx(v.length(), abs=5e-1):
                    found = True
            assert found

    def test_fft1d(self, setup_rlp):
        max_cell = 1.3 * max(setup_rlp["crystal_symmetry"].unit_cell().parameters()[:3])
        strategy = FFT1D(max_cell)
        basis_vectors, used = strategy.find_basis_vectors(setup_rlp["rlp"])
        self.check_results(setup_rlp["crystal_symmetry"].unit_cell(), basis_vectors)

    def test_fft3d(self, setup_rlp):
        max_cell = 1.3 * max(setup_rlp["crystal_symmetry"].unit_cell().parameters()[:3])
        strategy = FFT3D(max_cell)
        basis_vectors, used = strategy.find_basis_vectors(setup_rlp["rlp"])
        self.check_results(setup_rlp["crystal_symmetry"].unit_cell(), basis_vectors)

    def test_real_space_grid_search(self, setup_rlp):
        max_cell = 1.3 * max(setup_rlp["crystal_symmetry"].unit_cell().parameters()[:3])
        strategy = RealSpaceGridSearch(
            max_cell, target_unit_cell=setup_rlp["crystal_symmetry"].unit_cell()
        )
        basis_vectors, used = strategy.find_basis_vectors(setup_rlp["rlp"])
        self.check_results(setup_rlp["crystal_symmetry"].unit_cell(), basis_vectors)



from __future__ import annotations

import random
from math import pi

import pytest

from scitbx import matrix


def test_beam_parameters():
    from dxtbx.model import BeamFactory

    from dials.algorithms.refinement.parameterisation.beam_parameters import (
        BeamParameterisation,
    )
    from dials.algorithms.refinement.refinement_helpers import (
        get_fd_gradients,
        random_param_shift,
    )

    # make a random beam vector and parameterise it
    bf = BeamFactory()
    s0 = bf.make_beam(matrix.col.random(3, 0.5, 1.5), wavelength=1.2)
    s0p = BeamParameterisation(s0)

    # Let's do some basic tests. First, can we change parameter values and
    # update the modelled vector s0?
    s0_old = matrix.col(s0.get_s0())
    s0p.set_param_vals([1000 * 0.1, 1000 * 0.1, 0.8])
    assert matrix.col(s0.get_s0()).angle(s0_old) == pytest.approx(0.1413033, abs=1e-6)
    assert matrix.col(s0.get_s0()).length() == pytest.approx(0.8, abs=1e-6)

    # random initial orientations and wavelengths with a random parameter shifts
    attempts = 1000
    for i in range(attempts):

        # make a random beam vector and parameterise it
        sample_to_source = matrix.col.random(3, 0.5, 1.5).normalize()
        beam = bf.make_beam(sample_to_source, wavelength=random.uniform(0.8, 1.5))
        # Ensure consistent polarization (https://github.com/cctbx/dxtbx/issues/454)
        beam.set_polarization_normal(sample_to_source.ortho().normalize())

        s0p = BeamParameterisation(beam)

        # apply a random parameter shift
        p_vals = s0p.get_param_vals()
        p_vals = random_param_shift(p_vals, [1000 * pi / 9, 1000 * pi / 9, 0.01])
        s0p.set_param_vals(p_vals)

        # compare analytical and finite difference derivatives
        an_ds_dp = s0p.get_ds_dp()
        fd_ds_dp = get_fd_gradients(s0p, [1.0e-5 * pi / 180, 1.0e-5 * pi / 180, 1.0e-6])

        for j in range(3):
            try:
                assert list(fd_ds_dp[j] - an_ds_dp[j]) == pytest.approx(
                    (0, 0, 0), abs=1e-6
                )
            except Exception:
                print("for try", i)
                print("failure for parameter number", j)
                print("with fd_ds_dp = ")
                print(fd_ds_dp[j])
                print("and an_ds_dp = ")
                print(an_ds_dp[j])
                print("so that difference fd_ds_dp - an_ds_dp =")
                print(fd_ds_dp[j] - an_ds_dp[j])
                raise

        # Ensure the polarization normal vector remains orthogonal to the beam
        # (https://github.com/dials/dials/issues/1939)
        assert (
            abs(
                matrix.col(beam.get_unit_s0()).dot(
                    matrix.col(beam.get_polarization_normal())
                )
            )
            < 1e-10
        )
